# -*- coding: utf-8 -*-
"""The message module in agentscope."""

from ._message_block import (
    ContentBlock,
    TextBlock,
    ThinkingBlock,
    ToolUseBlock,
    ToolResultBlock,
    ImageBlock,
    AudioBlock,
    VideoBlock,
    Base64Source,
    URLSource,
)
from ._message_base import Msg


__all__ = [
    "TextBlock",
    "ThinkingBlock",
    "Base64Source",
    "URLSource",
    "ImageBlock",
    "AudioBlock",
    "VideoBlock",
    "ToolUseBlock",
    "ToolResultBlock",
    "ContentBlock",
    "Msg",
]



# -*- coding: utf-8 -*-
"""The common utilities for agentscope library."""
import asyncio
import base64
import functools
import inspect
import json
import os
import tempfile
import types
import typing
import uuid
from datetime import datetime
from typing import Union, Any, Callable, Type, Dict

import requests
from json_repair import repair_json
from pydantic import BaseModel

from .._logging import logger

if typing.TYPE_CHECKING:
    from mcp.types import Tool
else:
    Tool = "mcp.types.Tool"


def _json_loads_with_repair(
    json_str: str,
) -> Union[dict, list, str, float, int, bool, None]:
    """The given json_str maybe incomplete, e.g. '{"key', so we need to
    repair and load it into a Python object.
    """
    repaired = json_str
    try:
        repaired = repair_json(json_str)
    except Exception:
        pass

    try:
        return json.loads(repaired)
    except json.JSONDecodeError as e:
        raise ValueError(
            f"Failed to decode JSON string `{json_str}` after repairing it "
            f"into `{repaired}`. Error: {e}",
        ) from e


def _is_accessible_local_file(url: str) -> bool:
    """Check if the given URL is a local URL."""
    return os.path.isfile(url)


def _get_timestamp(add_random_suffix: bool = False) -> str:
    """Get the current timestamp in the format YYYY-MM-DD HH:MM:SS.sss."""
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

    if add_random_suffix:
        # Add a random suffix to the timestamp
        timestamp += f"_{os.urandom(3).hex()}"

    return timestamp


async def _is_async_func(func: Callable) -> bool:
    """Check if the given function is an async function, including
    coroutine functions, async generators, and coroutine objects.
    """

    return (
        inspect.iscoroutinefunction(func)
        or inspect.isasyncgenfunction(func)
        or isinstance(func, types.CoroutineType)
        or isinstance(func, types.GeneratorType)
        and asyncio.iscoroutine(func)
        or isinstance(func, functools.partial)
        and await _is_async_func(func.func)
    )


async def _execute_async_or_sync_func(
    func: Callable,
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Execute an async or sync function based on its type.

    Args:
        func (`Callable`):
            The function to be executed, which can be either async or sync.
        *args (`Any`):
            Positional arguments to be passed to the function.
        **kwargs (`Any`):
            Keyword arguments to be passed to the function.

    Returns:
        `Any`:
            The result of the function execution.
    """

    if await _is_async_func(func):
        return await func(*args, **kwargs)

    return func(*args, **kwargs)


def _get_bytes_from_web_url(
    url: str,
    max_retries: int = 3,
) -> str:
    """Get the bytes from a given URL.

    Args:
        url (`str`):
            The URL to fetch the bytes from.
        max_retries (`int`, defaults to `3`):
            The maximum number of retries.
    """
    for _ in range(max_retries):
        try:
            response = requests.get(url)
            response.raise_for_status()
            return response.content.decode("utf-8")

        except UnicodeDecodeError:
            return base64.b64encode(response.content).decode("ascii")

        except Exception as e:
            logger.info(
                "Failed to fetch bytes from URL %s. Error %s. Retrying...",
                url,
                str(e),
            )

    raise RuntimeError(
        f"Failed to fetch bytes from URL `{url}` after {max_retries} retries.",
    )


def _save_base64_data(
    media_type: str,
    base64_data: str,
) -> str:
    """Save the base64 data to a temp file and return the file path. The
    extension is guessed from the MIME type.

    Args:
        media_type (`str`):
            The MIME type of the data, e.g. "image/png", "audio/mpeg".
        base64_data (`str):
            The base64 data to be saved.
    """
    extension = "." + media_type.split("/")[-1]

    with tempfile.NamedTemporaryFile(
        suffix=f".{extension}",
        delete=False,
    ) as temp_file:
        decoded_data = base64.b64decode(base64_data)
        temp_file.write(decoded_data)
        temp_file.close()
        return temp_file.name


def _extract_json_schema_from_mcp_tool(tool: Tool) -> dict[str, Any]:
    """Extract JSON schema from MCP tool."""

    return {
        "type": "function",
        "function": {
            "name": tool.name,
            "description": tool.description,
            "parameters": {
                "type": "object",
                "properties": tool.inputSchema.get(
                    "properties",
                    {},
                ),
                "required": tool.inputSchema.get(
                    "required",
                    [],
                ),
            },
        },
    }


def _remove_title_field(schema: dict) -> None:
    """Remove the title field from the JSON schema to avoid
    misleading the LLM."""
    # The top level title field
    if "title" in schema:
        schema.pop("title")

    # properties
    if "properties" in schema:
        for prop in schema["properties"].values():
            if isinstance(prop, dict):
                _remove_title_field(prop)

    # items
    if "items" in schema and isinstance(schema["items"], dict):
        _remove_title_field(schema["items"])

    # additionalProperties
    if "additionalProperties" in schema and isinstance(
        schema["additionalProperties"],
        dict,
    ):
        _remove_title_field(
            schema["additionalProperties"],
        )


def _create_tool_from_base_model(
    structured_model: Type[BaseModel],
    tool_name: str = "generate_structured_output",
) -> Dict[str, Any]:
    """Create a function tool definition from a Pydantic BaseModel.
    This function converts a Pydantic BaseModel class into a tool definition
    that can be used with function calling API. The resulting tool
    definition includes the model's JSON schema as parameters, enabling
    structured output generation by forcing the model to call this function
    with properly formatted data.

    Args:
        structured_model (`Type[BaseModel]`):
            A Pydantic BaseModel class that defines the expected structure
            for the tool's output.
        tool_name (`str`, default `"generate_structured_output"`):
            The tool name that used to force the LLM to generate structured
            output by calling this function.

    Returns:
        `Dict[str, Any]`: A tool definition dictionary compatible with
            function calling API, containing type ("function") and
            function dictionary with name, description, and parameters
            (JSON schema).

    .. code-block:: python
        :caption: Example usage

        from pydantic import BaseModel

        class PersonInfo(BaseModel):
            name: str
            age: int
            email: str

        tool = _create_tool_from_base_model(PersonInfo, "extract_person")
        print(tool["function"]["name"])  # extract_person
        print(tool["type"])              # function

    .. note:: The function automatically removes the 'title' field from
        the JSON schema to ensure compatibility with function calling
        format. This is handled by the internal ``_remove_title_field()``
        function.
    """
    schema = structured_model.model_json_schema()

    _remove_title_field(schema)
    tool_definition = {
        "type": "function",
        "function": {
            "name": tool_name,
            "description": "Generate the required structured output with "
            "this function",
            "parameters": schema,
        },
    }
    return tool_definition


def _map_text_to_uuid(text: str) -> str:
    """Map the given text to a deterministic UUID string.

    Args:
        text (`str`):
            The input text to be mapped to a UUID.

    Returns:
        `str`:
            A deterministic UUID string derived from the input text.
    """
    return str(uuid.uuid3(uuid.NAMESPACE_DNS, text))



# -*- coding: utf-8 -*-
"""The models used in the plan module."""
from typing import Literal

import shortuuid
from pydantic import BaseModel, Field

from .._utils._common import _get_timestamp


class SubTask(BaseModel):
    """The subtask model used in the plan module."""

    name: str = Field(
        description=(
            "The subtask name, should be concise, descriptive and not"
            "exceed 10 words."
        ),
    )
    description: str = Field(
        description=(
            "The subtask description, including the constraints, target and "
            "outcome to be achieved. The description should be clear, "
            "specific and concise, and all the constraints, target and "
            "outcome should be specific and measurable."
        ),
    )
    expected_outcome: str = Field(
        description=(
            "The expected outcome of the subtask, which should be specific, "
            "concrete and measurable."
        ),
    )
    outcome: str | None = Field(
        description="The actual outcome of the subtask.",
        default=None,
    )
    state: Literal["todo", "in_progress", "done", "abandoned"] = Field(
        description="The state of the subtask.",
        default="todo",
    )
    created_at: str = Field(
        description="The time the subtask was created.",
        default_factory=_get_timestamp,
    )
    # Result related fields
    finished_at: str | None = Field(
        description="The time the subtask was finished.",
        default=None,
    )

    def finish(self, outcome: str) -> None:
        """Finish the subtask with the actual outcome."""
        self.state = "done"
        self.outcome = outcome
        self.finished_at = _get_timestamp()

    def to_oneline_markdown(self) -> str:
        """Convert the subtask to MarkDown format."""
        status_map = {
            "todo": "- []",
            "in_progress": "- [][WIP]",
            "done": "- [x]",
            "abandoned": "- [][Abandoned]",
        }
        return f"{status_map[self.state]} {self.name}"

    def to_markdown(self, detailed: bool = False) -> str:
        """Convert the subtask to MarkDown format.

        Args:
            detailed (`bool`, defaults to `False`):
                Whether to include detailed information about the subtask.
        """
        status_map = {
            "todo": "- [ ] ",
            "in_progress": "- [ ] [WIP]",
            "done": "- [x] ",
            "abandoned": "- [ ] [Abandoned]",
        }

        if detailed:
            markdown_strs = [
                f"{status_map[self.state]}{self.name}",
                f"\t- Created At: {self.created_at}",
                f"\t- Description: {self.description}",
                f"\t- Expected Outcome: {self.expected_outcome}",
                f"\t- State: {self.state}",
            ]

            if self.state == "done":
                markdown_strs.extend(
                    [
                        f"\t- Finished At: {self.finished_at}",
                        f"\t- Actual Outcome: {self.outcome}",
                    ],
                )

            return "\n".join(markdown_strs)

        return f"{status_map[self.state]}{self.name}"


class Plan(BaseModel):
    """The plan model used in the plan module, contains a list of subtasks."""

    id: str = Field(default_factory=shortuuid.uuid)
    name: str = Field(
        description=(
            "The plan name, should be concise, descriptive and not exceed 10 "
            "words."
        ),
    )
    description: str = Field(
        description=(
            "The plan description, including the constraints, target and "
            "outcome to be achieved. The description should be clear, "
            "specific and concise, and all the constraints, target and "
            "outcome should be specific and measurable."
        ),
    )
    expected_outcome: str = Field(
        description=(
            "The expected outcome of the plan, which should be specific, "
            "concrete and measurable."
        ),
    )
    subtasks: list[SubTask] = Field(
        description=("A list of subtasks that make up the plan."),
    )
    created_at: str = Field(
        description="The time the plan was created.",
        default_factory=_get_timestamp,
    )
    state: Literal["todo", "in_progress", "done", "abandoned"] = Field(
        description="The state of the plan.",
        default="todo",
    )
    finished_at: str | None = Field(
        description="The time the plan was finished.",
        default=None,
    )
    outcome: str | None = Field(
        description="The actual outcome of the plan.",
        default=None,
    )

    def refresh_plan_state(self) -> str:
        """Refresh the plan state based on the states of its subtasks. This
        function only switches the plan state between "todo" and "in_progress".

        # TODO: Handle the plan state much more formally.
        """
        if self.state in ["done", "abandoned"]:
            return ""

        any_in_progress = any(_.state == "in_progress" for _ in self.subtasks)

        if any_in_progress and self.state == "todo":
            self.state = "in_progress"
            return "The plan state has been updated to 'in_progress'."

        elif not any_in_progress and self.state == "in_progress":
            self.state = "todo"
            return "The plan state has been updated to 'todo'."

        return ""

    def finish(
        self,
        state: Literal["done", "abandoned"],
        outcome: str,
    ) -> None:
        """Finish the plan."""
        self.state = state
        self.outcome = outcome
        self.finished_at = _get_timestamp()

    def to_markdown(self, detailed: bool = False) -> str:
        """Convert the plan to MarkDown format."""
        subtasks_markdown = "\n".join(
            [
                subtask.to_markdown(
                    detailed=detailed,
                )
                for subtask in self.subtasks
            ],
        )

        return "\n".join(
            [
                f"# {self.name}",
                f"**Description**: {self.description}",
                f"**Expected Outcome**: {self.expected_outcome}",
                f"**State**: {self.state}",
                f"**Created At**: {self.created_at}",
                "## Subtasks",
                subtasks_markdown,
            ],
        )



# -*- coding: utf-8 -*-
"""The dashscope embedding module in agentscope."""
from datetime import datetime
from typing import Any, List, Literal

from ._cache_base import EmbeddingCacheBase
from ._embedding_response import EmbeddingResponse
from ._embedding_usage import EmbeddingUsage
from ._embedding_base import EmbeddingModelBase
from .._logging import logger
from ..message import TextBlock


class DashScopeTextEmbedding(EmbeddingModelBase):
    """DashScope text embedding API class.

    .. note:: From the `official documentation
    <https://bailian.console.aliyun.com/?tab=api#/api/?type=model&url=2712515>`_:

     - The max batch size that DashScope text embedding API
     supports is 10 for `text-embedding-v4` and `text-embedding-v3` models, and
     25 for `text-embedding-v2` and `text-embedding-v1` models.
     - The max token limit for a single input is 8192 tokens for `v4` and `v3`
     models, and 2048 tokens for `v2` and `v1` models.

    """

    supported_modalities: list[str] = ["text"]
    """This class only supports text input."""

    def __init__(
        self,
        api_key: str,
        model_name: str,
        dimensions: int = 1024,
        embedding_cache: EmbeddingCacheBase | None = None,
    ) -> None:
        """Initialize the DashScope text embedding model class.

        Args:
            api_key (`str`):
                The dashscope API key.
            model_name (`str`):
                The name of the embedding model.
            dimensions (`int`, defaults to 1024):
                The dimension of the embedding vector, refer to the
                `official documentation
                <https://bailian.console.aliyun.com/?tab=api#/api/?type=model&url=2712515>`_
                for more details.
            embedding_cache (`EmbeddingCacheBase`):
                The embedding cache class instance, used to cache the
                embedding results to avoid repeated API calls.
        """
        super().__init__(model_name, dimensions)

        self.api_key = api_key
        self.embedding_cache = embedding_cache
        self.batch_size_limit = 10

    async def _call_api(self, kwargs: dict[str, Any]) -> EmbeddingResponse:
        """Call the DashScope embedding API by the given keyword arguments."""

        if self.embedding_cache:
            cached_embeddings = await self.embedding_cache.retrieve(
                identifier=kwargs,
            )
            if cached_embeddings:
                return EmbeddingResponse(
                    embeddings=cached_embeddings,
                    usage=EmbeddingUsage(
                        tokens=0,
                        time=0,
                    ),
                    source="cache",
                )

        import dashscope

        start_time = datetime.now()
        response = dashscope.embeddings.TextEmbedding.call(
            api_key=self.api_key,
            **kwargs,
        )
        time = (datetime.now() - start_time).total_seconds()

        if response.status_code != 200:
            raise RuntimeError(
                f"Failed to get embedding from DashScope API: {response}",
            )

        if self.embedding_cache:
            await self.embedding_cache.store(
                identifier=kwargs,
                embeddings=[
                    _["embedding"] for _ in response.output["embeddings"]
                ],
            )

        return EmbeddingResponse(
            embeddings=[_["embedding"] for _ in response.output["embeddings"]],
            usage=EmbeddingUsage(
                tokens=response.usage["total_tokens"],
                time=time,
            ),
        )

    async def __call__(
        self,
        text: List[str | TextBlock],
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """Call the DashScope embedding API.

        Args:
            text (`List[str | TextBlock]`):
                The input text to be embedded. It can be a list of strings.
        """
        gather_text = []
        for _ in text:
            if isinstance(_, dict) and "text" in _:
                gather_text.append(_["text"])
            elif isinstance(_, str):
                gather_text.append(_)
            else:
                raise ValueError(
                    "Input text must be a list of strings or TextBlock dicts.",
                )

        if len(gather_text) > self.batch_size_limit:
            logger.info(
                "The input texts (%d) will be embedded with %d API calls due "
                f"to the batch size limit of {self.batch_size_limit} for "
                f"DashScope embedding API.",
                len(gather_text),
                (len(gather_text) + self.batch_size_limit - 1)
                // self.batch_size_limit,
            )

        # Handle the batch size limit for DashScope embedding API
        collected_embeddings = []
        collected_time = 0.0
        collected_tokens = 0
        collected_source: Literal["cache", "api"] = "cache"
        for _ in range(0, len(gather_text), self.batch_size_limit):
            batch_texts = gather_text[_ : _ + self.batch_size_limit]
            batch_kwargs = {
                "input": batch_texts,
                "model": self.model_name,
                "dimension": self.dimensions,
                **kwargs,
            }

            res = await self._call_api(batch_kwargs)

            collected_embeddings.extend(res.embeddings)
            collected_time += res.usage.time
            if res.usage.tokens:
                collected_tokens += res.usage.tokens
            if res.source == "api":
                collected_source = "api"

        return EmbeddingResponse(
            embeddings=collected_embeddings,
            usage=EmbeddingUsage(
                tokens=collected_tokens,
                time=collected_time,
            ),
            source=collected_source,
        )



# -*- coding: utf-8 -*-
"""The model response module."""

from dataclasses import dataclass, field
from typing import Literal, Sequence

from ._model_usage import ChatUsage
from .._utils._common import _get_timestamp
from .._utils._mixin import DictMixin
from ..message import (
    TextBlock,
    ToolUseBlock,
    ThinkingBlock,
    AudioBlock,
)
from ..types import JSONSerializableObject


@dataclass
class ChatResponse(DictMixin):
    """The response of chat models."""

    content: Sequence[TextBlock | ToolUseBlock | ThinkingBlock | AudioBlock]
    """The content of the chat response, which can include text blocks,
    tool use blocks, or thinking blocks."""

    id: str = field(default_factory=lambda: _get_timestamp(True))
    """The unique identifier formatter """

    created_at: str = field(default_factory=_get_timestamp)
    """When the response was created"""

    type: Literal["chat"] = field(default_factory=lambda: "chat")
    """The type of the response, which is always 'chat'."""

    usage: ChatUsage | None = field(default_factory=lambda: None)
    """The usage information of the chat response, if available."""

    metadata: dict[str, JSONSerializableObject] | None = field(
        default_factory=lambda: None,
    )
    """The metadata of the chat response"""



# -*- coding: utf-8 -*-
"""The OpenAI text embedding model class."""
from datetime import datetime
from typing import Any, List

from ._embedding_response import EmbeddingResponse
from ._embedding_usage import EmbeddingUsage
from ._cache_base import EmbeddingCacheBase
from ._embedding_base import EmbeddingModelBase
from ..message import TextBlock


class OpenAITextEmbedding(EmbeddingModelBase):
    """OpenAI text embedding model class."""

    supported_modalities: list[str] = ["text"]
    """This class only supports text input."""

    def __init__(
        self,
        api_key: str,
        model_name: str,
        dimensions: int = 1024,
        embedding_cache: EmbeddingCacheBase | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAI text embedding model class.

        Args:
            api_key (`str`):
                The OpenAI API key.
            model_name (`str`):
                The name of the embedding model.
            dimensions (`int`, defaults to 1024):
                The dimension of the embedding vector.
            embedding_cache (`EmbeddingCacheBase | None`, defaults to `None`):
                The embedding cache class instance, used to cache the
                embedding results to avoid repeated API calls.

        # TODO: handle batch size limit and token limit
        """
        import openai

        super().__init__(model_name, dimensions)

        self.client = openai.AsyncClient(api_key=api_key, **kwargs)
        self.embedding_cache = embedding_cache

    async def __call__(
        self,
        text: List[str | TextBlock],
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """Call the OpenAI embedding API.

        Args:
            text (`List[str | TextBlock]`):
                The input text to be embedded. It can be a list of strings.
        """
        gather_text = []
        for _ in text:
            if isinstance(_, dict) and "text" in _:
                gather_text.append(_["text"])
            elif isinstance(_, str):
                gather_text.append(_)
            else:
                raise ValueError(
                    "Input text must be a list of strings or TextBlock dicts.",
                )

        kwargs = {
            "input": gather_text,
            "model": self.model_name,
            "dimensions": self.dimensions,
            "encoding_format": "float",
            **kwargs,
        }

        if self.embedding_cache:
            cached_embeddings = await self.embedding_cache.retrieve(
                identifier=kwargs,
            )
            if cached_embeddings:
                return EmbeddingResponse(
                    embeddings=cached_embeddings,
                    usage=EmbeddingUsage(
                        tokens=0,
                        time=0,
                    ),
                    source="cache",
                )

        start_time = datetime.now()
        response = await self.client.embeddings.create(**kwargs)
        time = (datetime.now() - start_time).total_seconds()

        if self.embedding_cache:
            await self.embedding_cache.store(
                identifier=kwargs,
                embeddings=[_.embedding for _ in response.data],
            )

        return EmbeddingResponse(
            embeddings=[_.embedding for _ in response.data],
            usage=EmbeddingUsage(
                tokens=response.usage.total_tokens,
                time=time,
            ),
        )



# -*- coding: utf-8 -*-
"""Functional counterpart for Pipeline"""
import asyncio
from copy import deepcopy
from typing import Any, AsyncGenerator, Tuple, Coroutine
from ..agent import AgentBase
from ..message import Msg


async def sequential_pipeline(
    agents: list[AgentBase],
    msg: Msg | list[Msg] | None = None,
) -> Msg | list[Msg] | None:
    """An async syntactic sugar pipeline that executes a sequence of agents
    sequentially. The output of the previous agent will be passed as the
    input to the next agent. The final output will be the output of the
    last agent.

    Example:
        .. code-block:: python

            agent1 = ReActAgent(...)
            agent2 = ReActAgent(...)
            agent3 = ReActAgent(...)

            msg_input = Msg("user", "Hello", "user")

            msg_output = await sequential_pipeline(
                [agent1, agent2, agent3],
                msg_input
            )

    Args:
        agents (`list[AgentBase]`):
            A list of agents.
        msg (`Msg | list[Msg] | None`, defaults to `None`):
            The initial input that will be passed to the first agent.
    Returns:
        `Msg | list[Msg] | None`:
            The output of the last agent in the sequence.
    """
    for agent in agents:
        msg = await agent(msg)
    return msg


async def fanout_pipeline(
    agents: list[AgentBase],
    msg: Msg | list[Msg] | None = None,
    enable_gather: bool = True,
    **kwargs: Any,
) -> list[Msg]:
    """A fanout pipeline that distributes the same input to multiple agents.
    This pipeline sends the same message (or a deep copy of it) to all agents
    and collects their responses. Agents can be executed either concurrently
    using asyncio.gather() or sequentially depending on the enable_gather
    parameter.

    Example:
        .. code-block:: python

            agent1 = ReActAgent(...)
            agent2 = ReActAgent(...)
            agent3 = ReActAgent(...)

            msg_input = Msg("user", "Hello", "user")

            # Concurrent execution (default)
            results = await fanout_pipeline(
                [agent1, agent2, agent3],
                msg_input
            )

            # Sequential execution
            results = await fanout_pipeline(
                [agent1, agent2, agent3],
                msg_input,
                enable_gather=False
            )

    Args:
        agents (`list[AgentBase]`):
            A list of agents.
        msg (`Msg | list[Msg] | None`, defaults to `None`):
            The initial input that will be passed to all agents.
        enable_gather (`bool`, defaults to `True`):
            Whether to execute agents concurrently using `asyncio.gather()`.
            If False, agents are executed sequentially.
        **kwargs (`Any`):
            Additional keyword arguments passed to each agent during execution.

    Returns:
        `list[Msg]`:
            A list of response messages from each agent.
    """
    if enable_gather:
        tasks = [
            asyncio.create_task(agent(deepcopy(msg), **kwargs))
            for agent in agents
        ]

        return await asyncio.gather(*tasks)
    else:
        return [await agent(deepcopy(msg), **kwargs) for agent in agents]


async def stream_printing_messages(
    agents: list[AgentBase],
    coroutine_task: Coroutine,
    end_signal: str = "[END]",
) -> AsyncGenerator[Tuple[Msg, bool], None]:
    """This pipeline will gather the printing messages from agents when
    execute the given coroutine task, and yield them one by one.
    Only the messages that are printed by `await self.print(msg)` in the agent
    will be forwarded to the message queue and yielded by this pipeline.

    .. note:: The boolean in the yielded tuple indicates whether the message
     is the last **chunk** for a streaming message, not the last message
     returned by the agent. That means, there'll be multiple tuples with
     `is_last_chunk=True` if the agent prints multiple messages.

    .. note:: The messages with the same ``id`` is considered as the same
     message, e.g., the chunks of a streaming message.

    Args:
        agents (`list[AgentBase]`):
            A list of agents whose printing messages will be gathered and
            yielded.
        coroutine_task (`Coroutine`):
            The coroutine task to be executed. This task should involve the
            execution of the provided agents, so that their printing messages
            can be captured and yielded.
        end_signal (`str`, defaults to `"[END]"`):
            A special signal to indicate the end of message streaming. When
            this signal is received from the message queue, the generator will
            stop yielding messages and exit the loop.

    Returns:
        `AsyncGenerator[Tuple[Msg, bool], None]`:
            An async generator that yields tuples of (message, is_last_chunk).
            The `is_last_chunk` boolean indicates whether the message is the
            last chunk in a streaming message.
    """

    # Enable the message queue to get the intermediate messages
    queue = asyncio.Queue()
    for agent in agents:
        # Use one queue to gather messages from all agents
        agent.set_msg_queue_enabled(True, queue)

    # Execute the agent asynchronously
    task = asyncio.create_task(coroutine_task)

    if task.done():
        await queue.put(end_signal)
    else:
        task.add_done_callback(lambda _: queue.put_nowait(end_signal))

    # Receive the messages from the agent's message queue
    while True:
        # The message obj, and a boolean indicating whether it's the last chunk
        # in a streaming message
        printing_msg = await queue.get()

        # End the loop when the message is None
        if isinstance(printing_msg, str) and printing_msg == end_signal:
            break

        yield printing_msg



# -*- coding: utf-8 -*-
"""The tracing interface class in agentscope."""
from agentscope import _config


def setup_tracing(endpoint: str) -> None:
    """Set up the AgentScope tracing by configuring the endpoint URL.

    Args:
        endpoint (`str`):
            The endpoint URL for the tracing exporter.
    """
    # Lazy import
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
        OTLPSpanExporter,
    )
    from opentelemetry import trace

    tracer_provider = TracerProvider()
    exporter = OTLPSpanExporter(endpoint=endpoint)
    span_processor = BatchSpanProcessor(exporter)
    tracer_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(tracer_provider)

    _config.trace_enabled = True



# -*- coding: utf-8 -*-
"""The Chinese tools for ACEBench evaluation."""
from functools import wraps
from typing import Callable, Any

from ._ace_tools_api import (
    ReminderApi,
    FoodPlatformApi,
    TravelApi,
    MessageApi,
)
from ...message import TextBlock
from ...tool import ToolResponse


def _tool_function_wrapper(get_tool_function: Callable) -> Callable:
    """Wrap the tool function result to be ToolResponse."""

    @wraps(get_tool_function)
    def wrapper(self: "ACEPhone", name: str) -> Callable:
        """Wrap the tool function to return ToolResponse."""
        tool_function = get_tool_function(self, name)

        @wraps(tool_function)
        def wrapper_tool_function(*args: Any, **kwargs: Any) -> ToolResponse:
            """The wrapped tool function"""
            res = tool_function(*args, **kwargs)
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=str(res),
                    ),
                ],
            )

        return wrapper_tool_function

    return wrapper


class ACEPhone:
    """Simulate a user phone with various apps and functionalities in
    ACEBench. The code is implemented with reference to the
    `ACEBench <https://github.com/ACEBench/ACEBench>`_.
    """

    def __init__(self) -> None:
        """Initialize the shared state and apps for the ACEPhone."""
        self._state = {
            "wifi": False,
            "logged_in": False,
        }
        self._message_app = MessageApi(self._state)
        self._reminder_app = ReminderApi(self._state)
        self._food_platform_app = FoodPlatformApi(self._state)
        self._travel = TravelApi()

    def turn_on_wifi(self) -> dict[str, bool | str]:
        """开启WiFi连接。"""
        self._state["wifi"] = True
        return {"status": True, "message": "wifi已经打开"}

    def login_device(self) -> dict[str, bool | str]:
        """登录设备。"""
        self._state["logged_in"] = True
        return {"status": True, "message": "设备已经登录"}

    def load_initial_config(self, initial_config: dict) -> None:
        """Load the initial config from the application configuration."""
        # Empty initial config
        if len(initial_config) == 0:
            return

        # Fix the typo in ACEBench by renaming "Baspi" to "BaseApi"
        if "Baspi" in initial_config:
            initial_config["BaseApi"] = initial_config.pop("Baspi")

        # Verify state
        assert (
            "BaseApi" in initial_config
            and "wifi" in initial_config["BaseApi"]
            and "logged_in" in initial_config["BaseApi"]
        ), f"Invalid initial config: {initial_config}"

        self._state["wifi"] = initial_config["BaseApi"]["wifi"]
        self._state["logged_in"] = initial_config["BaseApi"]["logged_in"]

    def get_current_state(self) -> list[dict]:
        """Follow ACEBench to get the current state of the ACEPhone."""
        return [
            {"BaseApi": self._state},
            self._message_app.get_state_dict(),
            self._reminder_app.get_state_dict(),
            self._food_platform_app.get_state_dict(),
            self._travel.get_state_dict(),
        ]

    @_tool_function_wrapper
    def get_tool_function(self, name: str) -> Callable:
        """Get a tool function by name."""
        if name in [
            "turn_on_wifi",
            "login_device",
        ]:
            return getattr(self, name)

        if name in self._message_app.tool_functions:
            return getattr(self._message_app, name)

        if name in self._food_platform_app.tool_functions:
            return getattr(self._food_platform_app, name)

        if name in self._reminder_app.tool_functions:
            return getattr(self._reminder_app, name)

        if name in self._travel.tool_functions:
            return getattr(self._travel, name)

        raise ValueError(
            f"Tool function '{name}' not found in ACEPhone.",
        )



# -*- coding: utf-8 -*-
"""The data model for registered tool functions in AgentScope."""
from copy import deepcopy
from dataclasses import field, dataclass
from typing import Callable, Literal, Type, Awaitable

from pydantic import BaseModel

from ._response import ToolResponse
from .._utils._common import _remove_title_field
from ..message import ToolUseBlock
from ..types import ToolFunction, JSONSerializableObject


@dataclass
class RegisteredToolFunction:
    """The registered tool function class."""

    name: str
    """The name of the tool function."""
    group: str | Literal["basic"]
    """The belonging group of the tool function"""
    source: Literal["function", "mcp_server", "function_group"]
    """"The type of the tool function, can be `function` or `mcp_server`."""
    original_func: ToolFunction
    """The original function"""
    json_schema: dict
    """The JSON schema of the tool function, which is used to validate the """
    preset_kwargs: dict[str, JSONSerializableObject] = field(
        default_factory=dict,
    )
    """The preset keyword arguments, which won't be presented in the JSON
    schema and exposed to the user."""
    extended_model: Type[BaseModel] | None = None
    """The base model used to extend the JSON schema of the original tool
    function, so that we can dynamically adjust the tool function."""
    mcp_name: str | None = None
    """The name of the MCP, if the tool function comes from an MCP server."""
    postprocess_func: (
        Callable[
            [ToolUseBlock, ToolResponse],
            ToolResponse | None,
        ]
        | Callable[
            [ToolUseBlock, ToolResponse],
            Awaitable[ToolResponse | None],
        ]
    ) | None = None
    """The post-processing function that will be called after the tool
    function is executed, taking the tool call block and tool
    response as arguments. The function can be either sync or async. If it
    returns `None`, the tool result will be returned as is. If it returns a
    `ToolResponse`, the returned block will be used as the final tool
    response."""

    @property
    def extended_json_schema(self) -> dict:
        """Get the JSON schema of the tool function, if an extended model is
        set, the merged JSON schema will be returned."""
        if self.extended_model is None:
            return self.json_schema

        # Merge the extended model with the original JSON schema
        extended_schema = self.extended_model.model_json_schema()

        merged_schema = deepcopy(self.json_schema)

        _remove_title_field(  # pylint: disable=protected-access
            extended_schema,
        )

        for key, value in extended_schema["properties"].items():
            if key in self.json_schema["function"]["parameters"]["properties"]:
                raise ValueError(
                    f"The field `{key}` already exists in the original "
                    f"function schema of `{self.name}`. Try to use a "
                    "different name.",
                )

            merged_schema["function"]["parameters"]["properties"][key] = value

            if key in extended_schema.get("required", []):
                if "required" not in merged_schema["function"]["parameters"]:
                    merged_schema["function"]["parameters"]["required"] = []
                merged_schema["function"]["parameters"]["required"].append(key)
        return merged_schema



# -*- coding: utf-8 -*-
"""The Qdrant local vector store implementation."""
import json
from typing import Any, Literal, TYPE_CHECKING

from .._reader import Document
from ._store_base import VDBStoreBase
from .._document import DocMetadata
from ..._utils._common import _map_text_to_uuid
from ...types import Embedding

if TYPE_CHECKING:
    from qdrant_client import AsyncQdrantClient
else:
    AsyncQdrantClient = "qdrant_client.AsyncQdrantClient"


class QdrantStore(VDBStoreBase):
    """The Qdrant vector store implementation, supporting both local and
    remote Qdrant instances.

    .. note:: In Qdrant, we use the ``payload`` field to store the metadata,
    including the document ID, chunk ID, and original content.

    """

    def __init__(
        self,
        location: Literal[":memory:"] | str,
        collection_name: str,
        dimensions: int,
        distance: Literal["Cosine", "Euclid", "Dot", "Manhattan"] = "Cosine",
        client_kwargs: dict[str, Any] | None = None,
        collection_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the local Qdrant vector store.

        Args:
            location (`Literal[":memory:"] | str`):
                The location of the Qdrant instance. Use ":memory:" for
                in-memory Qdrant instance, or url for remote Qdrant instance,
                e.g. "http://localhost:6333" or a path to a directory.
            collection_name (`str`):
                The name of the collection to store the embeddings.
            dimensions (`int`):
                The dimension of the embeddings.
            distance (`Literal["Cosine", "Euclid", "Dot", "Manhattan"]`, \
            default to "Cosine"):
                The distance metric to use for the collection. Can be one of
                "Cosine", "Euclid", "Dot", or "Manhattan". Defaults to
                "Cosine".
            client_kwargs (`dict[str, Any] | None`, optional):
                Other keyword arguments for the Qdrant client.
            collection_kwargs (`dict[str, Any] | None`, optional):
                Other keyword arguments for creating the collection.
        """

        try:
            from qdrant_client import AsyncQdrantClient
        except ImportError as e:
            raise ImportError(
                "Qdrant client is not installed. Please install it with "
                "`pip install qdrant-client`.",
            ) from e

        client_kwargs = client_kwargs or {}
        self._client = AsyncQdrantClient(location=location, **client_kwargs)

        self.collection_name = collection_name
        self.dimensions = dimensions
        self.distance = distance
        self.collection_kwargs = collection_kwargs or {}

    async def _validate_collection(self) -> None:
        """Validate the collection exists, if not, create it."""
        if not await self._client.collection_exists(self.collection_name):
            from qdrant_client import models

            collections_kwargs = {
                "collection_name": self.collection_name,
                "vectors_config": models.VectorParams(
                    size=self.dimensions,
                    distance=getattr(models.Distance, self.distance.upper()),
                ),
                **self.collection_kwargs,
            }
            await self._client.create_collection(**collections_kwargs)

    async def add(self, documents: list[Document], **kwargs: Any) -> None:
        """Add embeddings to the Qdrant vector store.

        Args:
            documents (`list[Document]`):
                A list of embedding records to be recorded in the Qdrant store.
        """
        await self._validate_collection()

        from qdrant_client.models import PointStruct

        await self._client.upsert(
            collection_name=self.collection_name,
            points=[
                PointStruct(
                    id=_map_text_to_uuid(
                        json.dumps(
                            {
                                "doc_id": _.metadata.doc_id,
                                "chunk_id": _.metadata.chunk_id,
                                "content": _.metadata.content,
                            },
                            ensure_ascii=False,
                        ),
                    ),
                    vector=_.embedding,
                    payload=_.metadata,
                )
                for _ in documents
            ],
        )

    async def search(
        self,
        query_embedding: Embedding,
        limit: int,
        score_threshold: float | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """Search relevant documents from the Qdrant vector store.

        Args:
            query_embedding (`Embedding`):
                The embedding of the query text.
            limit (`int`):
                The number of relevant documents to retrieve.
            score_threshold (`float | None`, optional):
                The threshold of the score to filter the results.
            **kwargs (`Any`):
                Other keyword arguments for the Qdrant client search API.
        """
        res = await self._client.query_points(
            collection_name=self.collection_name,
            query=query_embedding,
            limit=limit,
            score_threshold=score_threshold,
            **kwargs,
        )

        collected_res = []
        for point in res.points:
            collected_res.append(
                Document(
                    embedding=point.vector,
                    score=point.score,
                    metadata=DocMetadata(**point.payload),
                ),
            )
        return collected_res

    async def delete(self, *args: Any, **kwargs: Any) -> None:
        """Delete is not implemented for QdrantStore."""
        raise NotImplementedError(
            "Delete is not implemented for QdrantStore.",
        )

    def get_client(self) -> AsyncQdrantClient:
        """Get the underlying Qdrant client, so that developers can access
        the full functionality of Qdrant.

        Returns:
            `AsyncQdrantClient`:
                The underlying Qdrant client.
        """
        return self._client



# -*- coding: utf-8 -*-
"""The JSON related types"""

from typing import Union

JSONPrimitive = Union[
    str,
    int,
    float,
    bool,
    None,
]

JSONSerializableObject = Union[
    JSONPrimitive,
    list["JSONSerializableObject"],
    dict[
        str,
        "JSONSerializableObject",
    ],
]



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""OpenAI Chat model class."""
from datetime import datetime
from typing import (
    Any,
    TYPE_CHECKING,
    List,
    AsyncGenerator,
    Literal,
    Type,
)
from collections import OrderedDict

from pydantic import BaseModel

from . import ChatResponse
from ._model_base import ChatModelBase
from ._model_usage import ChatUsage
from .._logging import logger
from .._utils._common import _json_loads_with_repair
from ..message import (
    ToolUseBlock,
    TextBlock,
    ThinkingBlock,
    AudioBlock,
    Base64Source,
)
from ..tracing import trace_llm
from ..types import JSONSerializableObject

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletion
    from openai import AsyncStream
else:
    ChatCompletion = "openai.types.chat.ChatCompletion"
    AsyncStream = "openai.types.chat.AsyncStream"


def _format_audio_data_for_qwen_omni(messages: list[dict]) -> None:
    """Qwen-omni uses OpenAI-compatible API but requires different audio
    data format than OpenAI with "data:;base64," prefix.
    Refer to `Qwen-omni documentation
    <https://bailian.console.aliyun.com/?tab=doc#/doc/?type=model&url=2867839>`_
    for more details.

    Args:
        messages (`list[dict]`):
            The list of message dictionaries from OpenAI formatter.
    """
    for msg in messages:
        if isinstance(msg.get("content"), list):
            for block in msg["content"]:
                if (
                    isinstance(block, dict)
                    and "input_audio" in block
                    and isinstance(block["input_audio"].get("data"), str)
                ):
                    if not block["input_audio"]["data"].startswith("http"):
                        block["input_audio"]["data"] = (
                            "data:;base64," + block["input_audio"]["data"]
                        )


class OpenAIChatModel(ChatModelBase):
    """The OpenAI chat model class."""

    def __init__(
        self,
        model_name: str,
        api_key: str | None = None,
        stream: bool = True,
        reasoning_effort: Literal["low", "medium", "high"] | None = None,
        organization: str = None,
        client_args: dict = None,
        generate_kwargs: dict[str, JSONSerializableObject] | None = None,
    ) -> None:
        """Initialize the openai client.

        Args:
            model_name (`str`, default `None`):
                The name of the model to use in OpenAI API.
            api_key (`str`, default `None`):
                The API key for OpenAI API. If not specified, it will
                be read from the environment variable `OPENAI_API_KEY`.
            stream (`bool`, default `True`):
                Whether to use streaming output or not.
            reasoning_effort (`Literal["low", "medium", "high"] | None`, \
            optional):
                Reasoning effort, supported for o3, o4, etc. Please refer to
                `OpenAI documentation
                <https://platform.openai.com/docs/guides/reasoning?api-mode=chat>`_
                for more details.
            organization (`str`, default `None`):
                The organization ID for OpenAI API. If not specified, it will
                be read from the environment variable `OPENAI_ORGANIZATION`.
            client_args (`dict`, default `None`):
                The extra keyword arguments to initialize the OpenAI client.
            generate_kwargs (`dict[str, JSONSerializableObject] | None`, \
             optional):
               The extra keyword arguments used in OpenAI API generation,
                e.g. `temperature`, `seed`.
        """

        super().__init__(model_name, stream)

        import openai

        self.client = openai.AsyncClient(
            api_key=api_key,
            organization=organization,
            **(client_args or {}),
        )

        self.reasoning_effort = reasoning_effort
        self.generate_kwargs = generate_kwargs or {}

    @trace_llm
    async def __call__(
        self,
        messages: list[dict],
        tools: list[dict] | None = None,
        tool_choice: Literal["auto", "none", "any", "required"]
        | str
        | None = None,
        structured_model: Type[BaseModel] | None = None,
        **kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        """Get the response from OpenAI chat completions API by the given
        arguments.

        Args:
            messages (`list[dict]`):
                A list of dictionaries, where `role` and `content` fields are
                required, and `name` field is optional.
            tools (`list[dict]`, default `None`):
                The tools JSON schemas that the model can use.
            tool_choice (`Literal["auto", "none", "any", "required"] | str \
            | None`, default `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", "any", "required", or specific tool
                 name. For more details, please refer to
                 https://platform.openai.com/docs/api-reference/responses/create#responses_create-tool_choice
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output. When provided, the model will be forced
                to return data that conforms to this schema by automatically
                converting the BaseModel to a tool function and setting
                `tool_choice` to enforce its usage. This enables structured
                output generation.

                .. note:: When `structured_model` is specified,
                    both `tools` and `tool_choice` parameters are ignored,
                    and the model will only perform structured output
                    generation without calling any other tools.

                For more details, please refer to the `official document
                <https://platform.openai.com/docs/guides/structured-outputs>`_

            **kwargs (`Any`):
                The keyword arguments for OpenAI chat completions API,
                e.g. `temperature`, `max_tokens`, `top_p`, etc. Please
                refer to the OpenAI API documentation for more details.

        Returns:
            `ChatResponse | AsyncGenerator[ChatResponse, None]`:
                The response from the OpenAI chat completions API.
        """

        # checking messages
        if not isinstance(messages, list):
            raise ValueError(
                "OpenAI `messages` field expected type `list`, "
                f"got `{type(messages)}` instead.",
            )
        if not all("role" in msg and "content" in msg for msg in messages):
            raise ValueError(
                "Each message in the 'messages' list must contain a 'role' "
                "and 'content' key for OpenAI API.",
            )

        # Qwen-omni requires different base64 audio format from openai
        if "omni" in self.model_name.lower():
            _format_audio_data_for_qwen_omni(messages)

        kwargs = {
            "model": self.model_name,
            "messages": messages,
            "stream": self.stream,
            **self.generate_kwargs,
            **kwargs,
        }
        if self.reasoning_effort and "reasoning_effort" not in kwargs:
            kwargs["reasoning_effort"] = self.reasoning_effort

        if tools:
            kwargs["tools"] = self._format_tools_json_schemas(tools)

        if tool_choice:
            self._validate_tool_choice(tool_choice, tools)
            kwargs["tool_choice"] = self._format_tool_choice(tool_choice)

        if self.stream:
            kwargs["stream_options"] = {"include_usage": True}

        start_datetime = datetime.now()

        if structured_model:
            if tools or tool_choice:
                logger.warning(
                    "structured_model is provided. Both 'tools' and "
                    "'tool_choice' parameters will be overridden and "
                    "ignored. The model will only perform structured output "
                    "generation without calling any other tools.",
                )
            kwargs.pop("stream", None)
            kwargs.pop("tools", None)
            kwargs.pop("tool_choice", None)
            kwargs["response_format"] = structured_model
            if not self.stream:
                response = await self.client.chat.completions.parse(**kwargs)
            else:
                response = self.client.chat.completions.stream(**kwargs)
                return self._parse_openai_stream_response(
                    start_datetime,
                    response,
                    structured_model,
                )
        else:
            response = await self.client.chat.completions.create(**kwargs)

        if self.stream:
            return self._parse_openai_stream_response(
                start_datetime,
                response,
                structured_model,
            )

        # Non-streaming response
        parsed_response = self._parse_openai_completion_response(
            start_datetime,
            response,
            structured_model,
        )

        return parsed_response

    async def _parse_openai_stream_response(
        self,
        start_datetime: datetime,
        response: AsyncStream,
        structured_model: Type[BaseModel] | None = None,
    ) -> AsyncGenerator[ChatResponse, None]:
        """Given an OpenAI streaming completion response, extract the content
         blocks and usages from it and yield ChatResponse objects.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`AsyncStream`):
                OpenAI AsyncStream object to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            `AsyncGenerator[ChatResponse, None]`:
                An async generator that yields ChatResponse objects containing
                the content blocks and usage information for each chunk in
                the streaming response.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        usage, res = None, None
        text = ""
        thinking = ""
        audio = ""
        tool_calls = OrderedDict()
        metadata: dict | None = None
        contents: List[
            TextBlock | ToolUseBlock | ThinkingBlock | AudioBlock
        ] = []

        async with response as stream:
            async for item in stream:
                if structured_model:
                    if item.type != "chunk":
                        continue
                    chunk = item.chunk
                else:
                    chunk = item

                if chunk.usage:
                    usage = ChatUsage(
                        input_tokens=chunk.usage.prompt_tokens,
                        output_tokens=chunk.usage.completion_tokens,
                        time=(datetime.now() - start_datetime).total_seconds(),
                    )

                if not chunk.choices:
                    if usage and contents:
                        res = ChatResponse(
                            content=contents,
                            usage=usage,
                            metadata=metadata,
                        )
                        yield res
                    continue

                choice = chunk.choices[0]

                thinking += (
                    getattr(choice.delta, "reasoning_content", None) or ""
                )
                text += choice.delta.content or ""

                if (
                    hasattr(choice.delta, "audio")
                    and "data" in choice.delta.audio
                ):
                    audio += choice.delta.audio["data"]
                if (
                    hasattr(choice.delta, "audio")
                    and "transcript" in choice.delta.audio
                ):
                    text += choice.delta.audio["transcript"]

                for tool_call in choice.delta.tool_calls or []:
                    if tool_call.index in tool_calls:
                        if tool_call.function.arguments is not None:
                            tool_calls[tool_call.index][
                                "input"
                            ] += tool_call.function.arguments

                    else:
                        tool_calls[tool_call.index] = {
                            "type": "tool_use",
                            "id": tool_call.id,
                            "name": tool_call.function.name,
                            "input": tool_call.function.arguments or "",
                        }

                contents = []

                if thinking:
                    contents.append(
                        ThinkingBlock(
                            type="thinking",
                            thinking=thinking,
                        ),
                    )

                if audio:
                    media_type = self.generate_kwargs.get("audio", {}).get(
                        "format",
                        "wav",
                    )
                    contents.append(
                        AudioBlock(
                            type="audio",
                            source=Base64Source(
                                data=audio,
                                media_type=f"audio/{media_type}",
                                type="base64",
                            ),
                        ),
                    )

                if text:
                    contents.append(
                        TextBlock(
                            type="text",
                            text=text,
                        ),
                    )

                    if structured_model:
                        metadata = _json_loads_with_repair(text)

                for tool_call in tool_calls.values():
                    contents.append(
                        ToolUseBlock(
                            type=tool_call["type"],
                            id=tool_call["id"],
                            name=tool_call["name"],
                            input=_json_loads_with_repair(
                                tool_call["input"] or "{}",
                            ),
                        ),
                    )

                if not contents:
                    continue

                res = ChatResponse(
                    content=contents,
                    usage=usage,
                    metadata=metadata,
                )
                yield res

    def _parse_openai_completion_response(
        self,
        start_datetime: datetime,
        response: ChatCompletion,
        structured_model: Type[BaseModel] | None = None,
    ) -> ChatResponse:
        """Given an OpenAI chat completion response object, extract the content
            blocks and usages from it.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`ChatCompletion`):
                OpenAI ChatCompletion object to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            ChatResponse (`ChatResponse`):
                A ChatResponse object containing the content blocks and usage.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        content_blocks: List[
            TextBlock | ToolUseBlock | ThinkingBlock | AudioBlock
        ] = []
        metadata: dict | None = None

        if response.choices:
            choice = response.choices[0]
            if (
                hasattr(choice.message, "reasoning_content")
                and choice.message.reasoning_content is not None
            ):
                content_blocks.append(
                    ThinkingBlock(
                        type="thinking",
                        thinking=response.choices[0].message.reasoning_content,
                    ),
                )

            if choice.message.content:
                content_blocks.append(
                    TextBlock(
                        type="text",
                        text=response.choices[0].message.content,
                    ),
                )
            if choice.message.audio:
                media_type = self.generate_kwargs.get("audio", {}).get(
                    "format",
                    "mp3",
                )
                content_blocks.append(
                    AudioBlock(
                        type="audio",
                        source=Base64Source(
                            data=choice.message.audio.data,
                            media_type=f"audio/{media_type}",
                            type="base64",
                        ),
                    ),
                )

                if choice.message.audio.transcript:
                    content_blocks.append(
                        TextBlock(
                            type="text",
                            text=choice.message.audio.transcript,
                        ),
                    )

            for tool_call in choice.message.tool_calls or []:
                content_blocks.append(
                    ToolUseBlock(
                        type="tool_use",
                        id=tool_call.id,
                        name=tool_call.function.name,
                        input=_json_loads_with_repair(
                            tool_call.function.arguments,
                        ),
                    ),
                )

            if structured_model:
                metadata = choice.message.parsed.model_dump()

        usage = None
        if response.usage:
            usage = ChatUsage(
                input_tokens=response.usage.prompt_tokens,
                output_tokens=response.usage.completion_tokens,
                time=(datetime.now() - start_datetime).total_seconds(),
            )

        parsed_response = ChatResponse(
            content=content_blocks,
            usage=usage,
            metadata=metadata,
        )

        return parsed_response

    def _format_tools_json_schemas(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schemas to the OpenAI format."""
        return schemas

    def _format_tool_choice(
        self,
        tool_choice: Literal["auto", "none", "any", "required"] | str | None,
    ) -> str | dict | None:
        """Format tool_choice parameter for API compatibility.

        Args:
            tool_choice (`Literal["auto", "none", "any", "required"] | str \
            | None`, default `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", "any", "required", or specific tool
                 name. For more details, please refer to
                 https://platform.openai.com/docs/api-reference/responses/create#responses_create-tool_choice
        Returns:
            `dict | None`:
                The formatted tool choice configuration dict, or None if
                    tool_choice is None.
        """
        if tool_choice is None:
            return None
        mode_mapping = {
            "auto": "auto",
            "none": "none",
            "any": "required",
            "required": "required",
        }
        if tool_choice in mode_mapping:
            return mode_mapping[tool_choice]
        return {"type": "function", "function": {"name": tool_choice}}



# -*- coding: utf-8 -*-
# type: ignore
# pylint: disable=too-many-lines
# pylint: disable=too-many-statements
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-return-statements
"""The travel API for the ACEBench simulation tools in AgentScope."""

from datetime import datetime, timedelta


class TravelApi:
    """旅行预订系统类。

    提供航班查询、用户认证、预订管理等功能的旅行系统。
    支持直飞和中转航班查询、航班预订、预订修改和取消等功能。
    """

    tool_functions: list[str] = [
        "get_user_details",
        "get_flight_details",
        "get_reservation_details",
        "reserve_flight",
        "cancel_reservation",
        "modify_flight",
    ]

    def __init__(self) -> None:
        """初始化旅行系统。

        设置用户档案和航班信息，包含用户信息、航班数据和预订记录。
        """
        # 初始化用户信息
        self.users = {
            "user1": {
                "user_name": "Eve",
                "password": "password123",
                "cash_balance": 2000.0,
                "bank_balance": 50000.0,
                "membership_level": "regular",
            },
            "user2": {
                "user_name": "Frank",
                "password": "password456",
                "cash_balance": 8000.0,
                "bank_balance": 8000.0,
                "membership_level": "silver",
            },
            "user3": {
                "user_name": "Grace",
                "password": "password789",
                "cash_balance": 1000.0,
                "bank_balance": 5000.0,
                "membership_level": "gold",
            },
        }

        # 初始化航班信息
        self.flights = [
            {
                "flight_no": "CA1234",
                "origin": "北京",
                "destination": "上海",
                "depart_time": "2024-07-15 08:00:00",
                "arrival_time": "2024-07-15 10:30:00",
                "status": "available",
                "seats_available": 5,
                "economy_price": 1200,
                "business_price": 3000,
            },
            {
                "flight_no": "MU5678",
                "origin": "上海",
                "destination": "北京",
                "depart_time": "2024-07-16 09:00:00",
                "arrival_time": "2024-07-16 11:30:00",
                "status": "available",
                "seats_available": 3,
                "economy_price": 1900,
                "business_price": 3000,
            },
            {
                "flight_no": "CZ4321",
                "origin": "上海",
                "destination": "北京",
                "depart_time": "2024-07-16 20:00:00",
                "arrival_time": "2024-07-16 22:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 2500,
                "business_price": 4000,
            },
            {
                "flight_no": "CZ4352",
                "origin": "上海",
                "destination": "北京",
                "depart_time": "2024-07-17 20:00:00",
                "arrival_time": "2024-07-17 22:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1600,
                "business_price": 2500,
            },
            {
                "flight_no": "MU3561",
                "origin": "北京",
                "destination": "南京",
                "depart_time": "2024-07-18 08:00:00",
                "arrival_time": "2024-07-18 10:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 4000,
            },
            {
                "flight_no": "MU1566",
                "origin": "北京",
                "destination": "南京",
                "depart_time": "2024-07-18 20:00:00",
                "arrival_time": "2024-07-18 22:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 4000,
            },
            {
                "flight_no": "CZ1765",
                "origin": "南京",
                "destination": "深圳",
                "depart_time": "2024-07-17 20:30:00",
                "arrival_time": "2024-07-17 22:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 2500,
            },
            {
                "flight_no": "CZ1765",
                "origin": "南京",
                "destination": "深圳",
                "depart_time": "2024-07-18 12:30:00",
                "arrival_time": "2024-07-18 15:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 2500,
            },
            {
                "flight_no": "MH1765",
                "origin": "厦门",
                "destination": "成都",
                "depart_time": "2024-07-17 12:30:00",
                "arrival_time": "2024-07-17 15:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 2500,
            },
            {
                "flight_no": "MH2616",
                "origin": "成都",
                "destination": "厦门",
                "depart_time": "2024-07-18 18:30:00",
                "arrival_time": "2024-07-18 21:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 2500,
            },
            {
                "flight_no": "MH2616",
                "origin": "成都",
                "destination": "福州",
                "depart_time": "2024-07-16 18:30:00",
                "arrival_time": "2024-07-16 21:00:00",
                "status": "available",
                "seats_available": 8,
                "economy_price": 1500,
                "business_price": 2500,
            },
        ]

        # 初始化预订列表
        self.reservations = [
            {
                "reservation_id": "res_1",
                "user_id": "user1",
                "flight_no": "CA1234",
                "payment_method": "bank",
                "cabin": "经济舱",
                "baggage": 1,
                "origin": "北京",
                "destination": "上海",
            },
            {
                "reservation_id": "res_2",
                "user_id": "user1",
                "flight_no": "MU5678",
                "payment_method": "bank",
                "cabin": "商务舱",
                "baggage": 1,
                "origin": "上海",
                "destination": "北京",
            },
            {
                "reservation_id": "res_3",
                "user_id": "user2",
                "flight_no": "MH1765",
                "payment_method": "bank",
                "cabin": "商务舱",
                "baggage": 1,
                "origin": "厦门",
                "destination": "成都",
            },
            {
                "reservation_id": "res_4",
                "user_id": "user2",
                "flight_no": "MU2616",
                "payment_method": "bank",
                "cabin": "商务舱",
                "baggage": 1,
                "origin": "成都",
                "destination": "厦门",
            },
        ]

    def get_state_dict(self) -> dict:
        """Get the current state dict of the TravelApi."""
        return {
            "Travel": {
                "users": self.users,
                "reservations": self.reservations,
            },
        }

    # 根据出发地和到达地查询航班

    def get_flight_details(
        self,
        origin: str = None,
        destination: str = None,
    ) -> list[dict] | str:
        """根据出发地和到达地查询航班的基本信息。

        Args:
            origin (str, optional): 出发地城市名称。默认为None。
            destination (str, optional): 目的地城市名称。默认为None。

        Returns:
            list[dict] | str: 符合条件的航班列表或无航班的提示信息。
        """
        flights = self.flights

        # 过滤出发地
        if origin:
            flights = [
                flight for flight in flights if flight["origin"] == origin
            ]

        # 过滤到达地
        if destination:
            flights = [
                flight
                for flight in flights
                if flight["destination"] == destination
            ]
        if len(flights) == 0:
            return "没有符合条件的直达航班"
        # 返回查询结果
        return [
            {
                "flight_no": flight["flight_no"],
                "origin": flight["origin"],
                "destination": flight["destination"],
                "depart_time": flight["depart_time"],
                "arrival_time": flight["arrival_time"],
                "status": flight["status"],
                "seats_available": flight["seats_available"],
                "economy_price": flight["economy_price"],
                "business_price": flight["business_price"],
            }
            for flight in flights
        ]

    def get_user_details(self, user_id: str, password: str) -> dict:
        """根据用户名和密码查询用户信息。

        Args:
            user_id (str): 用户ID。
            password (str): 用户密码。

        Returns:
            dict: 用户信息字典（不包含密码）或错误信息。
        """
        user = self.users.get(user_id)
        if user and user["password"] == password:
            return {
                key: value for key, value in user.items() if key != "password"
            }
        return {"status": "error", "message": "用户名或密码不正确"}

    def get_reservation_details(
        self,
        reservation_id: str = None,
        user_id: str = None,
    ) -> list[dict] | dict:
        """根据预订ID或用户ID查询预订信息，包括对应航班的基本信息。

        Args:
            reservation_id (str, optional): 预订ID。默认为None。
            user_id (str, optional): 用户ID。默认为None。

        Returns:
            `list[dict] | dict`:
                详细预订信息列表或错误信息字典。
        """
        # 根据预订ID或用户ID筛选预订信息
        if reservation_id:
            reservations = [
                reservation
                for reservation in self.reservations
                if reservation["reservation_id"] == reservation_id
            ]
        elif user_id:
            reservations = [
                reservation
                for reservation in self.reservations
                if reservation["user_id"] == user_id
            ]
        else:
            return {"status": "error", "message": "请提供有效的预订ID或用户ID"}

        # 对每个预订，附加航班信息
        detailed_reservations = []
        for reservation in reservations:
            flight_info = next(
                (
                    flight
                    for flight in self.flights
                    if flight["flight_no"] == reservation["flight_no"]
                ),
                None,
            )
            detailed_reservation = {**reservation, "flight_info": flight_info}
            detailed_reservations.append(detailed_reservation)

        return detailed_reservations

    def authenticate_user(self, user_id: str, password: str) -> dict:
        """验证用户身份。

        Args:
            user_id (str): 用户ID。
            password (str): 用户密码。

        Returns:
            `dict`:
                用户信息字典或错误信息字典。
        """
        user = self.users.get(user_id)
        if user and user["password"] == password:
            return user
        return {"status": "error", "message": "用户名或密码不正确"}

    def get_baggage_allowance(
        self,
        membership_level: str,
        cabin_class: str,
    ) -> int:
        """获取用户基于会员等级和舱位的免费托运行李限额。

        Args:
            membership_level (str): 会员等级 ("regular", "silver", "gold")。
            cabin_class (str): 舱位 ("基础经济舱", "经济舱", "商务舱")。

        Returns:
            int: 免费托运行李数量。
        """
        allowance = {
            "regular": {"经济舱": 1, "商务舱": 2},
            "silver": {"经济舱": 2, "商务舱": 3},
            "gold": {"经济舱": 3, "商务舱": 3},
        }
        return allowance.get(membership_level, {}).get(cabin_class, 0)

    def find_transfer_flights(
        self,
        origin_city: str,
        transfer_city: str,
        destination_city: str,
    ) -> list[dict] | str:
        """查找从出发城市到目的地城市的中转航班。

        确保第一班航班降落时间早于第二班航班起飞时间。

        Args:
            origin_city (str): 出发城市。
            transfer_city (str): 中转城市。
            destination_city (str): 到达城市。

        Returns:
            list[dict] | str:
                满足条件的中转航班列表，每个航班包含两段航程的信息，或无航班提示。
        """
        # 获取从出发城市到中转城市的航班
        first_leg_flights: list[dict] = [
            flight
            for flight in self.flights
            if flight["origin"] == origin_city
            and flight["destination"] == transfer_city
            and flight["status"] == "available"
        ]

        # 获取从中转城市到目的地城市的航班
        second_leg_flights = [
            flight
            for flight in self.flights
            if flight["origin"] == transfer_city
            and flight["destination"] == destination_city
            and flight["status"] == "available"
        ]

        # 存储符合条件的中转航班
        transfer_flights = []

        # 遍历第一段航班和第二段航班，查找符合时间条件的组合
        for first_flight in first_leg_flights:
            first_arrival = datetime.strptime(
                first_flight["arrival_time"],
                "%Y-%m-%d %H:%M:%S",
            )

            for second_flight in second_leg_flights:
                second_departure = datetime.strptime(
                    str(second_flight["depart_time"]),
                    "%Y-%m-%d %H:%M:%S",
                )

                # 检查第一班航班降落时间早于第二班航班起飞时间
                if first_arrival < second_departure:
                    transfer_flights.append(
                        {
                            "first_leg": first_flight,
                            "second_leg": second_flight,
                        },
                    )

        # 返回符合条件的中转航班列表
        if transfer_flights:
            return transfer_flights
        else:
            return "未找到符合条件的中转航班。"

    def calculate_baggage_fee(
        self,
        membership_level: str,
        cabin_class: str,
        baggage_count: int,
    ) -> float:
        """计算行李费用。

        Args:
            membership_level (str): 会员等级。
            cabin_class (str): 舱位等级。
            baggage_count (int): 行李数量。

        Returns:
            float: 额外行李费用。
        """
        free_baggage = {
            "regular": {"经济舱": 1, "商务舱": 2},
            "silver": {"经济舱": 2, "商务舱": 3},
            "gold": {"经济舱": 3, "商务舱": 3},
        }
        free_limit = free_baggage[membership_level][cabin_class]
        additional_baggage = max(baggage_count - free_limit, 0)
        return additional_baggage * 50

    def update_balance(
        self,
        user: dict,
        payment_method: str,
        amount: float,
    ) -> bool:
        """更新用户的余额。

        Args:
            user (dict): 用户信息字典。
            payment_method (str): 支付方式（"cash" 或 "bank"）。
            amount (float): 更新金额（正数表示增加，负数表示减少）。

        Returns:
            bool: 如果余额充足且更新成功，返回 True，否则返回 False。
        """
        if payment_method == "cash":
            if user["cash_balance"] + amount < 0:
                return False  # 余额不足
            user["cash_balance"] += amount
        elif payment_method == "bank":
            if user["bank_balance"] + amount < 0:
                return False  # 余额不足
            user["bank_balance"] += amount
        return True

    def reserve_flight(
        self,
        user_id: str,
        password: str,
        flight_no: str,
        cabin: str,
        payment_method: str,
        baggage_count: int,
    ) -> str:
        """预订航班。

        Args:
            user_id (str): 用户ID。
            password (str): 用户密码。
            flight_no (str): 航班号。
            cabin (str): 舱位等级。
            payment_method (str): 支付方式。
            baggage_count (int): 行李数量。

        Returns:
            str: 预订结果信息。
        """
        user = self.authenticate_user(user_id, password)
        if not user:
            return "认证失败，请检查用户ID和密码。"

        # 检查航班和座位
        flight = next(
            (
                f
                for f in self.flights
                if f["flight_no"] == flight_no and f["status"] == "available"
            ),
            None,
        )

        # 计算航班价格
        price: int = (
            flight["economy_price"]
            if cabin == "经济舱"
            else flight["business_price"]
        )
        total_cost = price

        # 计算行李费用
        baggage_fee = self.calculate_baggage_fee(
            user["membership_level"],
            cabin,
            baggage_count,
        )
        total_cost += baggage_fee

        # 检查支付方式
        if payment_method not in ["cash", "bank"]:
            return "支付方式无效"

        # 更新预定后的余额
        if payment_method == "cash":
            if total_cost > self.users.get(user_id)["cash_balance"]:
                return "cash余额不足，请考虑换一种支付方式"
            self.users.get(user_id)["cash_balance"] -= total_cost
        else:
            if total_cost > self.users.get(user_id)["bank_balance"]:
                return "bank余额不足，请考虑换一种支付方式"
            self.users.get(user_id)["bank_balance"] -= total_cost

        # 更新航班信息并生成预订
        flight["seats_available"] -= 1
        reservation_id = f"res_{len(self.reservations) + 1}"
        reservation = {
            "reservation_id": reservation_id,
            "user_id": user_id,
            "flight_no": flight_no,
            "payment_method": payment_method,
            "cabin": cabin,
            "baggage": baggage_count,
        }
        self.reservations.append(reservation)

        return f"预订成功，预订号：{reservation_id}，" f"总费用：{total_cost}元（包含行李费用）。"

    def modify_flight(
        self,
        user_id: str,
        reservation_id: str,
        new_flight_no: str = None,
        new_cabin: str = None,
        add_baggage: int = 0,
        new_payment_method: str = None,
    ) -> str:
        """修改航班预订，包括更改航班、舱位和行李。

        Args:
            user_id (str): 用户ID。
            reservation_id (str): 预订ID。
            new_flight_no (str, optional): 新的航班号。默认为None。
            new_cabin (str, optional): 新的舱位。默认为None。
            add_baggage (int, optional): 新增托运行李的数量。默认为0。
            new_payment_method (str, optional): 新的付款方式。默认为None。

        Returns:
            str: 修改结果信息。
        """
        # 获取对应的预订
        reservation = next(
            (
                r
                for r in self.reservations
                if r["reservation_id"] == reservation_id
                and r["user_id"] == user_id
            ),
            None,
        )
        if not reservation:
            return "预订未找到或用户ID不匹配。"

        # 检查当前预订的航班信息
        current_flight = next(
            (
                f
                for f in self.flights
                if f["flight_no"] == reservation["flight_no"]
            ),
            None,
        )
        if not current_flight:
            return "航班信息未找到。"

        # 获取原始支付方式或新提供的支付方式
        payment_method = (
            new_payment_method
            if new_payment_method
            else reservation["payment_method"]
        )
        user = self.users[user_id]
        if not user:
            return "用户信息未找到。"

        # 存储处理结果
        result_messages = []

        if new_flight_no and new_flight_no != reservation["flight_no"]:
            # 更新航班号（若提供）但必须匹配出发地和目的地
            new_flight = next(
                (f for f in self.flights if f["flight_no"] == new_flight_no),
                None,
            )
            if (
                new_flight
                and new_flight["origin"] == current_flight["origin"]
                and new_flight["destination"] == current_flight["destination"]
            ):
                reservation["flight_no"] = new_flight_no
                result_messages.append("航班号已更改。")
            else:
                return "航班更改失败：新的航班号无效或目的地不匹配。"

        # 更新舱位（若提供）并计算价格差价
        if new_cabin and new_cabin != reservation.get("cabin"):
            price_difference = self.calculate_price_difference(
                current_flight,
                reservation["cabin"],
                new_cabin,
            )
            reservation["cabin"] = new_cabin
            if price_difference > 0:
                # 扣除差价
                if self.update_balance(
                    user,
                    payment_method,
                    -price_difference,
                ):
                    result_messages.append(
                        f"舱位更改成功。已支付差价: {price_difference}。",
                    )
                else:
                    result_messages.append("余额不足，无法支付舱位差价。")
            elif price_difference < 0:
                # 退款
                self.update_balance(user, payment_method, -price_difference)
                result_messages.append(f"舱位更改成功。已退款差价: {-price_difference}。")

        # 增加托运行李，检查免费限额和计算费用
        if add_baggage > 0:
            membership = user["membership_level"]
            max_free_baggage = self.get_baggage_allowance(
                membership,
                reservation["cabin"],
            )
            current_baggage = reservation.get("baggage", 0)
            total_baggage = current_baggage + add_baggage
            extra_baggage = max(0, total_baggage - max_free_baggage)
            baggage_cost = extra_baggage * 50
            if baggage_cost > 0:
                # 扣除行李费用
                if self.update_balance(user, payment_method, -baggage_cost):
                    result_messages.append(
                        f"行李已增加。需支付额外费用: {baggage_cost}。",
                    )
                else:
                    result_messages.append("余额不足，无法支付额外行李费用。")
            reservation["baggage"] = total_baggage

        # 返回最终结果
        if not result_messages:
            result_messages.append("修改完成，无需额外费用。")
        return " ".join(result_messages)

    def cancel_reservation(
        self,
        user_id: str,
        reservation_id: str,
        reason: str,
    ) -> str:
        """取消预订。

        Args:
            user_id (str): 用户ID。
            reservation_id (str): 预订ID。
            reason (str): 取消原因。

        Returns:
            str: 取消结果信息。
        """
        # 设置默认当前时间为 2024年7月14日早上6点
        current_time = datetime(2024, 7, 14, 6, 0, 0)

        # 验证用户和预订是否存在
        user = self.users.get(user_id, None)
        if not user:
            return "用户ID无效。"

        reservation = next(
            (
                r
                for r in self.reservations
                if r["reservation_id"] == reservation_id
                and r["user_id"] == user_id
            ),
            None,
        )
        if not reservation:
            return "预订ID无效或与该用户无关。"

        # 检查航班信息是否存在
        flight = next(
            (
                f
                for f in self.flights
                if f["flight_no"] == reservation["flight_no"]
            ),
            None,
        )
        if not flight:
            return "航班信息无效。"

        # 检查航班是否已起飞
        depart_time = datetime.strptime(
            flight["depart_time"],
            "%Y-%m-%d %H:%M:%S",
        )
        if current_time > depart_time:
            return "航段已使用，无法取消。"

        # 计算距离出发时间
        time_until_departure = depart_time - current_time
        cancel_fee = 0
        refund_amount = 0

        # 获取航班价格
        flight_price = (
            flight["economy_price"]
            if reservation["cabin"] == "经济舱"
            else flight["business_price"]
        )

        # 取消政策及退款计算
        if reason == "航空公司取消航班":
            # 航空公司取消航班，全额退款
            refund_amount = flight_price
            self.process_refund(user, refund_amount)
            return f"航班已取消，您的预订将被免费取消，已退款{refund_amount}元。"

        elif time_until_departure > timedelta(days=1):
            # 离出发时间超过24小时免费取消
            refund_amount = flight_price
            self.process_refund(user, refund_amount)
            return f"距离出发时间超过24小时，免费取消成功，已退款{refund_amount}元。"

        else:
            # 若不符合免费取消条件，可根据需求设置取消费
            cancel_fee = flight_price * 0.1  # 假设取消费为票价的10%
            refund_amount = flight_price - cancel_fee
            self.process_refund(user, refund_amount)
            return f"距离出发时间不足24小时，已扣除取消费{cancel_fee}元，退款{refund_amount}元。"

    def process_refund(self, user: dict, amount: float) -> str:
        """将退款金额添加到用户的现金余额中。

        Args:
            user (dict): 用户信息字典。
            amount (float): 退款金额。
        """
        user["cash_balance"] += amount
        return f"已成功处理退款，{user['user_name']}的现金余额增加了{amount}元。"

    def calculate_price_difference(
        self,
        flight: dict,
        old_cabin: str,
        new_cabin: str,
    ) -> float:
        """计算舱位价格差异。

        Args:
            flight (dict): 航班信息字典。
            old_cabin (str): 原舱位等级。
            new_cabin (str): 新舱位等级。

        Returns:
            float: 价格差异（正数表示需支付差价，负数表示退款）。
        """
        cabin_prices = {
            "经济舱": flight["economy_price"],
            "商务舱": flight["business_price"],
        }
        old_price = cabin_prices.get(old_cabin, 0)
        new_price = cabin_prices.get(new_cabin, 0)
        return new_price - old_price



# -*- coding: utf-8 -*-
"""The evaluator storage base class for storing solution and evaluation
results."""
from abc import abstractmethod
from typing import Any, Callable

from .._metric_base import MetricResult
from .._solution import SolutionOutput
from ...agent import AgentBase


class EvaluatorStorageBase:
    """Used to store the solution results and evaluation results to support
    resuming the evaluation process"""

    @abstractmethod
    def save_solution_result(
        self,
        task_id: str,
        repeat_id: str,
        output: SolutionOutput,
        **kwargs: Any,
    ) -> None:
        """Save the solution result.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            output (`SolutionOutput`):
                The solution output to be saved.
        """

    @abstractmethod
    def get_evaluation_result(
        self,
        task_id: str,
        repeat_id: str,
        metric_name: str,
    ) -> MetricResult:
        """Get the evaluation result by the given task id and repeat id

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            metric_name (`str`):
                The metric name.

        Returns:
            `MetricResult`:
                The evaluation result for the given task and repeat ID.
        """

    @abstractmethod
    def save_evaluation_result(
        self,
        task_id: str,
        repeat_id: str,
        evaluation: MetricResult,
        **kwargs: Any,
    ) -> None:
        """Save the evaluation result.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            evaluation (`MetricResult`):
                The evaluation result to be saved.
        """

    @abstractmethod
    def get_solution_result(
        self,
        task_id: str,
        repeat_id: str,
        **kwargs: Any,
    ) -> SolutionOutput:
        """Get the solution result for the given task and repeat id.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Returns:
            `SolutionOutput`:
                The solution output for the given task and repeat ID.
        """

    @abstractmethod
    def solution_result_exists(self, task_id: str, repeat_id: str) -> bool:
        """Check if the solution for the given task and repeat is finished.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Returns:
            `bool`:
                True if the solution result file exists, False otherwise.
        """

    @abstractmethod
    def evaluation_result_exists(
        self,
        task_id: str,
        repeat_id: str,
        metric_name: str,
    ) -> bool:
        """Check if the evaluation result for the given solution and metric
        is finished.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            metric_name (`str`):
                The name of the metric.

        Returns:
            `bool`:
                True if the evaluation result file exists, False otherwise.
        """

    @abstractmethod
    def save_aggregation_result(
        self,
        aggregation_result: dict,
        **kwargs: Any,
    ) -> None:
        """Save the aggregation result.

        Args:
            aggregation_result (`dict`):
                A dictionary containing the aggregation result.
        """

    @abstractmethod
    def aggregation_result_exists(
        self,
        **kwargs: Any,
    ) -> bool:
        """Check if the aggregation result exists

        Returns:
            `bool`:
                `True` if the aggregation result file exists.
        """

    @abstractmethod
    def save_evaluation_meta(self, meta_info: dict) -> None:
        """Save the evaluation meta information.

        Args:
            meta_info (`dict`):
                A dictionary containing the meta information.
        """

    @abstractmethod
    def get_agent_pre_print_hook(
        self,
        task_id: str,
        repeat_id: str,
    ) -> Callable[[AgentBase, dict], None]:
        """Get a pre-print hook function for the agent to save the agent
        printing in the evaluation storage.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Returns:
            `Callable[[AgentBase, dict], None]`:
                A hook function that takes an `AgentBase` instance and a
                keyword arguments dictionary as input, saving the agent's
                printing Msg into the evaluation storage.
        """



# -*- coding: utf-8 -*-
"""The retrieval-augmented generation (RAG) module in AgentScope."""

from ._document import (
    DocMetadata,
    Document,
)
from ._reader import (
    ReaderBase,
    TextReader,
    PDFReader,
    ImageReader,
    WordReader,
)
from ._store import (
    VDBStoreBase,
    QdrantStore,
    MilvusLiteStore,
)
from ._knowledge_base import KnowledgeBase
from ._simple_knowledge import SimpleKnowledge


__all__ = [
    "ReaderBase",
    "TextReader",
    "PDFReader",
    "ImageReader",
    "WordReader",
    "DocMetadata",
    "Document",
    "VDBStoreBase",
    "QdrantStore",
    "MilvusLiteStore",
    "KnowledgeBase",
    "SimpleKnowledge",
]



# -*- coding: utf-8 -*-
"""The Anthropic token counter class."""
from typing import Any


class AnthropicTokenCounter:
    """The Anthropic token counter class."""

    def __init__(self, model_name: str, api_key: str, **kwargs: Any) -> None:
        """Initialize the Anthropic token counter.

        Args:
            model_name (`str`):
                The name of the Anthropic model to use, e.g. "claude-2".
            api_key (`str`):
                The API key for Anthropic.
        """
        import anthropic

        self.client = anthropic.AsyncAnthropic(api_key=api_key, **kwargs)
        self.model_name = model_name

    async def count(
        self,
        messages: list[dict],
        tools: list[dict] | None = None,
        **kwargs: Any,
    ) -> int:
        """Count the number of tokens for the given messages

        .. note:: The Anthropic token counting API requires the multimodal
         data to be in base64 format,

        Args:
            messages (`list[dict]`):
                A list of dictionaries, where `role` and `content` fields are
                required.
            tools (`list[dict] | None`, defaults to `None`):
                The tools JSON schemas that the model can use.
            **kwargs (`Any`):
                Additional keyword arguments for the token counting API.
        """
        system_message = None
        if messages[0].get("role") == "system":
            system_message = messages.pop(0)

        extra_kwargs: dict = {
            "model": self.model_name,
            "messages": messages,
            **kwargs,
        }

        if tools:
            extra_kwargs["tools"] = tools

        if system_message:
            extra_kwargs["system"] = system_message

        res = await self.client.messages.count_tokens(**extra_kwargs)

        return res.input_tokens



# -*- coding: utf-8 -*-
"""The base class for evaluator in evaluation."""
import collections
import json
from abc import abstractmethod
from typing import Callable, Coroutine, Any

from .._solution import SolutionOutput
from .._task import Task
from .._benchmark_base import BenchmarkBase
from .._evaluator_storage import EvaluatorStorageBase
from .._metric_base import MetricType
from ..._utils._common import _get_timestamp


class EvaluatorBase:
    """The class that runs the evaluation process."""

    def __init__(
        self,
        name: str,
        benchmark: BenchmarkBase,
        n_repeat: int,
        storage: EvaluatorStorageBase,
    ) -> None:
        """Initialize the evaluator.

        Args:
            name (`str`):
                The name of this evaluator.
            benchmark: (`BenchmarkBase`):
                A benchmark instance inheriting from `BenchmarkBase` that
                defines the evaluation dataset.
            n_repeat (`int`):
                How many times to repeat the evaluation for each task.
            storage (`EvaluatorStorageBase`):
                A instance inheriting from the child class of
                `EvaluatorStorageBase` that supports storing and loading
                solution output and evaluation results.
        """
        self.name = name
        self.benchmark = benchmark
        self.n_repeat = n_repeat
        self.storage = storage

    @abstractmethod
    async def run(
        self,
        solution: Callable[
            [Task, Callable],
            Coroutine[Any, Any, SolutionOutput],
        ],
    ) -> None:
        """Run the evaluation and return the results.

        Args:
            solution (`Callable[[Task, Callable], Coroutine[Any, Any, \
            SolutionOutput]]`):
                A async function that takes a `Task` instance and a pre-hook
                as input and returns a `SolutionOutput` instance.
        """

    async def _save_evaluation_meta(self) -> None:
        """Save the evaluation meta information."""
        self.storage.save_evaluation_meta(
            {
                "evaluation_name": self.name,
                "created_at": _get_timestamp(),
                "total_repeats": self.n_repeat,
                "benchmark": {
                    "name": self.benchmark.name,
                    "description": self.benchmark.description,
                    "total_tasks": len(self.benchmark),
                },
                "schema_version": 1,
            },
        )

    async def aggregate(self) -> None:  # pylint: disable=too-many-branches
        """Aggregate the evaluation results and save an overall result."""
        meta_info: dict = {
            "total_tasks": len(self.benchmark),
            "total_repeats": self.n_repeat,
            "repeats": {},
            "schema_version": 1,
        }

        for repeat_index in range(self.n_repeat):
            repeat_id = str(repeat_index)
            current_repeat: dict = {
                "completed_tasks": 0,
                "incomplete_tasks": 0,
                "metrics": {},
                "completed_ids": [],
                "incomplete_ids": [],
            }
            for task in self.benchmark:
                for metric in task.metrics:
                    # Create a new dict in aggregated_result
                    if metric.name not in current_repeat["metrics"]:
                        current_repeat["metrics"][metric.name] = {
                            "type": metric.metric_type,
                            "involved_tasks": 0,
                            "completed_tasks": 0,
                            "incomplete_tasks": 0,
                            "aggregation": {},
                            "distribution": collections.defaultdict(list),
                        }

                    # Record the submitted task
                    current_repeat["metrics"][metric.name][
                        "involved_tasks"
                    ] += 1

                    # Not finished
                    if not self.storage.evaluation_result_exists(
                        task.id,
                        repeat_id,
                        metric.name,
                    ):
                        if task.id not in current_repeat["incomplete_ids"]:
                            current_repeat["incomplete_tasks"] += 1
                            current_repeat["incomplete_ids"].append(task.id)
                        current_repeat["metrics"][metric.name][
                            "incomplete_tasks"
                        ] += 1
                        continue

                    if task.id not in current_repeat["completed_ids"]:
                        current_repeat["completed_tasks"] += 1
                        current_repeat["completed_ids"].append(task.id)
                    current_repeat["metrics"][metric.name][
                        "completed_tasks"
                    ] += 1

                    # Get the evaluation result
                    eval_result = self.storage.get_evaluation_result(
                        task.id,
                        repeat_id,
                        metric.name,
                    )

                    # Record the metric result
                    if metric.metric_type == MetricType.CATEGORY:
                        current_repeat["metrics"][metric.name]["distribution"][
                            eval_result.result
                        ].append(
                            task.id,
                        )

                    elif metric.metric_type == MetricType.NUMERICAL:
                        current_repeat["metrics"][metric.name]["distribution"][
                            task.id
                        ] = eval_result.result

            print("Repeat ID:", repeat_id)

            for metric, value in current_repeat["metrics"].items():
                print("\tMetric:", metric)
                print("\t\tType:", value["type"])
                print("\t\tInvolved tasks:", value["involved_tasks"])
                print("\t\tCompleted tasks:", value["completed_tasks"])
                print("\t\tIncomplete tasks:", value["incomplete_tasks"])

                if value["type"] == MetricType.CATEGORY:
                    # Count the distribution
                    for category, task_ids in value["distribution"].items():
                        value["aggregation"][category] = (
                            len(task_ids) * 1.0 / value["involved_tasks"]
                        )

                elif value["type"] == MetricType.NUMERICAL:
                    scores = list(value["distribution"].values())
                    value["aggregation"] = {
                        "mean": sum(scores) / value["involved_tasks"],
                        "max": max(scores),
                        "min": min(scores),
                    }

                print(
                    "\t\tAggregation:",
                    json.dumps(
                        value["aggregation"],
                        indent=4,
                        ensure_ascii=False,
                    ).replace("\n", "\n\t\t"),
                )

            meta_info["repeats"][repeat_id] = current_repeat

        # save
        self.storage.save_aggregation_result(meta_info)



# -*- coding: utf-8 -*-
"""The logger for agentscope."""

import logging


_DEFAULT_FORMAT = (
    "%(asctime)s | %(levelname)-7s | "
    "%(module)s:%(funcName)s:%(lineno)s - %(message)s"
)

logger = logging.getLogger("as")


def setup_logger(
    level: str,
    filepath: str | None = None,
) -> None:
    """Set up the agentscope logger.

    Args:
        level (`str`):
            The logging level, chosen from "INFO", "DEBUG", "WARNING",
            "ERROR", "CRITICAL".
        filepath (`str | None`, optional):
            The filepath to save the logging output.
    """
    if level not in ["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"]:
        raise ValueError(
            f"Invalid logging level: {level}. Must be one of "
            f"'INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL'.",
        )
    logger.handlers.clear()
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(_DEFAULT_FORMAT))
    logger.addHandler(handler)

    if filepath:
        handler = logging.FileHandler(filepath)
        handler.setFormatter(logging.Formatter(_DEFAULT_FORMAT))
        logger.addHandler(handler)

    logger.propagate = False


setup_logger("INFO")



# -*- coding: utf-8 -*-
"""Workflow for agent learning."""

from typing import (
    Dict,
    Callable,
    Awaitable,
    get_type_hints,
)

import inspect

from .._logging import logger
from ..model import TrinityChatModel


WorkflowType = Callable[[Dict, TrinityChatModel], Awaitable[float]]


def _validate_function_signature(func: Callable) -> bool:
    """Validate if a function matches the workflow type signature.

    Args:
        func (Callable): The function to validate.
    """
    # check if the function is asynchronous
    if not inspect.iscoroutinefunction(func):
        logger.warning("The function is not asynchronous.")
        return False
    # Define expected parameter types and return type manually
    expected_params = [
        ("task", Dict),
        ("model", TrinityChatModel),
    ]
    expected_return = float

    func_signature = inspect.signature(func)
    func_hints = get_type_hints(func)

    # Check if the number of parameters matches
    if len(func_signature.parameters) != len(expected_params):
        logger.warning(
            "Expected %d parameters, but got %d",
            len(expected_params),
            len(func_signature.parameters),
        )
        return False

    # Validate each parameter's name and type
    for (param_name, _), (expected_name, expected_type) in zip(
        func_signature.parameters.items(),
        expected_params,
    ):
        if (
            param_name != expected_name
            or func_hints.get(param_name) != expected_type
        ):
            logger.warning(
                "Expected parameter %s of type %s, but got %s of type %s",
                expected_name,
                expected_type,
                param_name,
                func_hints.get(param_name),
            )
            return False

    # Validate the return type
    return_annotation = func_hints.get("return", None)
    if return_annotation != expected_return:
        logger.warning(
            "Expected return type %s, but got %s",
            expected_return,
            return_annotation,
        )
        return False

    return True



# -*- coding: utf-8 -*-
"""The ACE benchmark class in agentscope. The code is implemented with
reference to the `ACEBench <https://github.com/ACEBench/ACEBench>`_
under the MIT license."""
import json
import os
from typing import Generator

import json5
import requests
from tqdm import tqdm

from ._ace_metric import ACEAccuracy, ACEProcessAccuracy
from ._ace_tools_zh import ACEPhone
from .._benchmark_base import BenchmarkBase
from .._task import Task


class ACEBenchmark(BenchmarkBase):
    """The ACE benchmark for evaluating AI agents."""

    data_dir_url: str = (
        "https://raw.githubusercontent.com/ACEBench/ACEBench/main/data_all"
    )
    """The URL to the data dir"""

    data_subdir: list[str] = [
        # "data_en",  # TODO: enable English version
        "data_zh",
    ]

    ground_truth_dir: str = "possible_answer"

    data_files: list[str] = [
        "data_agent_multi_step.json",
        "data_agent_multi_turn.json",
        # "data_normal_atom_bool.json",
        # "data_normal_atom_enum.json",
        # "data_normal_atom_list.json",
        # "data_normal_atom_number.json",
        # "data_normal_atom_object_deep.json",
        # "data_normal_atom_object_short.json",
        #
        # "data_normal_multi_turn_user_adjust.json",
        # "data_normal_multi_turn_user_switch.json",
        #
        # "data_normal_preference.json",
        # "data_normal_similar_api.json",
        # "data_normal_single_turn_parallel_function.json",
        # "data_normal_single_turn_single_function.json",
        #
        # "data_special_error_param.json",
        # "data_special_incomplete.json",
        # "data_special_irrelevant.json",
    ]
    """The data filenames"""

    def __init__(
        self,
        data_dir: str,
    ) -> None:
        """Initialize the ACEBenchmark

        Args:
            data_dir (`str`):
                The directory where the dataset is downloaded and saved.
        """
        super().__init__(
            name="ACEBench",
            description="The ACE benchmark for evaluating AI agents.",
        )

        self.data_dir = os.path.abspath(data_dir)

        if os.path.exists(data_dir) and not os.path.isdir(data_dir):
            raise RuntimeError(
                f"The data_dir `{data_dir}` is not a valid directory path.",
            )

        os.makedirs(data_dir, exist_ok=True)

        if not self._verify_data():
            self._download_data()

        self.dataset = self._load_data()

    def _load_data(self) -> list[dict]:
        """Load the dataset from the data directory."""
        dataset = []
        for subdir in self.data_subdir:
            for filename in self.data_files:
                file_path = os.path.join(self.data_dir, subdir, filename)

                gt_path = os.path.join(
                    self.data_dir,
                    subdir,
                    self.ground_truth_dir,
                    filename,
                )
                gt_dataset = {}
                with open(gt_path, "r", encoding="utf-8") as gt_file:
                    for line in gt_file:
                        gt_data = json5.loads(line)
                        gt_dataset[gt_data["id"]] = gt_data

                with open(file_path, "r", encoding="utf-8") as f:
                    for line in f:
                        data = json5.loads(line)
                        gt = gt_dataset[data["id"]]
                        gt.pop("id", None)
                        data["ground_truth"] = gt["ground_truth"]
                        data["mile_stone"] = gt["mile_stone"]
                        data["language"] = subdir.rsplit(
                            "_",
                            maxsplit=1,
                        )[-1]
                        data["tags"] = {
                            "language": data["language"],
                            "category": filename.split(
                                ".",
                                maxsplit=1,
                            )[0].removeprefix(
                                "data_",
                            ),
                        }
                        dataset.append(data)

        return dataset

    def _verify_data(self) -> bool:
        """Verify the data completeness and integrity."""
        for subdir in self.data_subdir:
            for filename in self.data_files:
                file_path = os.path.join(self.data_dir, subdir, filename)
                if not os.path.exists(file_path):
                    return False

                gt_path = os.path.join(
                    self.data_dir,
                    subdir,
                    self.ground_truth_dir,
                    filename,
                )
                if not os.path.exists(gt_path):
                    return False

        return True

    def _download_data(self) -> None:
        """Download the data from the URL"""
        for subdir in self.data_subdir:
            subdir_path = os.path.join(self.data_dir, subdir)
            subdir_gt_path = os.path.join(subdir_path, self.ground_truth_dir)
            os.makedirs(subdir_path, exist_ok=True)
            os.makedirs(subdir_gt_path, exist_ok=True)
            for filename in tqdm(
                self.data_files,
                desc=f"Downloading {subdir}",
            ):
                response = requests.get(
                    f"{self.data_dir_url}/{subdir}/{filename}",
                )
                response.raise_for_status()
                with open(os.path.join(subdir_path, filename), "wb") as f:
                    f.write(response.content)

                gt_response = requests.get(
                    f"{self.data_dir_url}/{subdir}/"
                    f"{self.ground_truth_dir}/{filename}",
                )
                gt_response.raise_for_status()
                with open(os.path.join(subdir_gt_path, filename), "wb") as f:
                    f.write(gt_response.content)

    @staticmethod
    def _data_to_task(item: dict) -> Task:
        """Convert a dataset item to a Task object."""
        # Start the simulated phone and load initial configuration
        ace_phone = ACEPhone()
        ace_phone.load_initial_config(item["initial_config"])

        # Obtain tool functions
        tools: list[tuple] = []
        for function_schema in item["function"]:
            name = function_schema["name"]

            # Handle the schema differences
            formatted_schema = json.loads(
                json.dumps(
                    function_schema,
                ).replace(
                    '"type": "dict"',
                    '"type": "object"',
                ),
            )

            tool_function = ace_phone.get_tool_function(name)
            tools.append(
                (
                    tool_function,
                    {
                        "type": "function",
                        "function": formatted_schema,
                    },
                ),
            )

        return Task(
            id=item["id"],
            input=item["question"],
            ground_truth={
                "state": item["ground_truth"],
                "mile_stone": item.get("mile_stone", []),
            },
            tags=item.get("tags", {}),
            metrics=[
                ACEAccuracy(item["ground_truth"]),
                ACEProcessAccuracy(item["mile_stone"]),
            ],
            metadata={
                # The phone is used to extract the final state after finishing
                # the task.
                "phone": ace_phone,
                # The provided tools for this task, used to equip the agent
                "tools": tools,
            },
        )

    def __iter__(self) -> Generator[Task, None, None]:
        """Iterate over the benchmark."""
        for item in self.dataset:
            yield self._data_to_task(item)

    def __getitem__(self, index: int) -> Task:
        """Get a task by index."""
        return self._data_to_task(self.dataset[index])

    def __len__(self) -> int:
        """Get the length of the benchmark."""
        return len(self.dataset)



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""The Python code execution tool in agentscope."""

import asyncio
import os
import sys
import tempfile
from typing import Any

import shortuuid

from ...message import TextBlock
from .._response import ToolResponse


async def execute_python_code(
    code: str,
    timeout: float = 300,
    **kwargs: Any,
) -> ToolResponse:
    """Execute the given python code in a temp file and capture the return
    code, standard output and error. Note you must `print` the output to get
    the result, and the tmp file will be removed right after the execution.

    Args:
        code (`str`):
            The Python code to be executed.
        timeout (`float`, defaults to `300`):
            The maximum time (in seconds) allowed for the code to run.

    Returns:
        `ToolResponse`:
            The response containing the return code, standard output, and
            standard error of the executed code.
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = os.path.join(temp_dir, f"tmp_{shortuuid.uuid()}.py")
        with open(temp_file, "w", encoding="utf-8") as f:
            f.write(code)

        proc = await asyncio.create_subprocess_exec(
            sys.executable,
            "-u",
            temp_file,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        try:
            await asyncio.wait_for(proc.wait(), timeout=timeout)
            stdout, stderr = await proc.communicate()
            stdout_str = stdout.decode("utf-8")
            stderr_str = stderr.decode("utf-8")
            returncode = proc.returncode

        except asyncio.TimeoutError:
            stderr_suffix = (
                f"TimeoutError: The code execution exceeded "
                f"the timeout of {timeout} seconds."
            )
            returncode = -1
            try:
                proc.terminate()
                stdout, stderr = await proc.communicate()
                stdout_str = stdout.decode("utf-8")
                stderr_str = stderr.decode("utf-8")
                if stderr_str:
                    stderr_str += f"\n{stderr_suffix}"
                else:
                    stderr_str = stderr_suffix
            except ProcessLookupError:
                stdout_str = ""
                stderr_str = stderr_suffix

        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text=f"<returncode>{returncode}</returncode>"
                    f"<stdout>{stdout_str}</stdout>"
                    f"<stderr>{stderr_str}</stderr>",
                ),
            ],
        )



# -*- coding: utf-8 -*-
"""Pipeline classes."""
from typing import Any

from ._functional import sequential_pipeline, fanout_pipeline
from ..agent import AgentBase
from ..message import Msg


class SequentialPipeline:
    """An async sequential pipeline class, which executes a sequence of
    agents sequentially. Compared with functional pipeline, this class
    can be re-used."""

    def __init__(
        self,
        agents: list[AgentBase],
    ) -> None:
        """Initialize a sequential pipeline class

        Args:
            agents (`list[AgentBase]`):
                A list of agents.
        """
        self.agents = agents

    async def __call__(
        self,
        msg: Msg | list[Msg] | None = None,
    ) -> Msg | list[Msg] | None:
        """Execute the sequential pipeline

        Args:
            msg (`Msg | list[Msg] | None`, defaults to `None`):
                The initial input that will be passed to the first agent.
        """
        return await sequential_pipeline(
            agents=self.agents,
            msg=msg,
        )


class FanoutPipeline:
    """An async fanout pipeline class, which distributes the same input to
    multiple agents. Compared with functional pipeline, this class can be
    re-used and configured with default parameters."""

    def __init__(
        self,
        agents: list[AgentBase],
        enable_gather: bool = True,
    ) -> None:
        """Initialize a fanout pipeline class

        Args:
            agents (`list[AgentBase]`):
                A list of agents to execute.
            enable_gather (`bool`, defaults to `True`):
                Whether to execute agents concurrently
                using `asyncio.gather()`. If False, agents are executed
                sequentially.
        """
        self.agents = agents
        self.enable_gather = enable_gather

    async def __call__(
        self,
        msg: Msg | list[Msg] | None = None,
        **kwargs: Any,
    ) -> list[Msg]:
        """Execute the fanout pipeline

        Args:
            msg (`Msg | list[Msg] | None`, defaults to `None`):
                The input message that will be distributed to all agents.
            **kwargs (`Any`):
                Additional keyword arguments passed to each agent during
                execution.

        Returns:
            `list[Msg]`:
                A list of output messages from all agents.
        """

        return await fanout_pipeline(
            agents=self.agents,
            msg=msg,
            enable_gather=self.enable_gather,
            **kwargs,
        )



# -*- coding: utf-8 -*-
"""MsgHub is designed to share messages among a group of agents."""
from typing import Any

import shortuuid

from .._logging import logger

from ..agent import AgentBase
from ..message import Msg


class MsgHub:
    """MsgHub class that controls the subscription of the participated agents.

    Example:
        In the following example, the reply message from `agent1`, `agent2`,
        and `agent3` will be broadcasted to all the other agents in the MsgHub.

        .. code-block:: python

            with MsgHub(participant=[agent1, agent2, agent3]):
                agent1()
                agent2()

        Actually, it has the same effect as the following code, but much more
        easy and elegant!

        .. code-block:: python

            x1 = agent1()
            agent2.observe(x1)
            agent3.observe(x1)

            x2 = agent2()
            agent1.observe(x2)
            agent3.observe(x2)

    """

    def __init__(
        self,
        participants: list[AgentBase],
        announcement: list[Msg] | Msg | None = None,
        enable_auto_broadcast: bool = True,
        name: str | None = None,
    ) -> None:
        """Initialize a MsgHub context manager.

        Args:
            participants (`list[AgentBase]`):
                A list of agents that participate in the MsgHub.
            announcement （`list[Msg] | Msg | None`):
                The message that will be broadcast to all participants when
                entering the MsgHub.
            enable_auto_broadcast (`bool`, defaults to `True`):
                Whether to enable automatic broadcasting of the replied
                message from any participant to all other participants. If
                disabled, the MsgHub will only serve as a manual message
                broadcaster with the `announcement` argument and the
                `broadcast()` method.
            name (`str | None`):
                The name of this MsgHub. If not provided, a random ID
                will be generated.
        """
        self.name = name or shortuuid.uuid()
        self.participants = participants
        self.announcement = announcement
        self.enable_auto_broadcast = enable_auto_broadcast

    async def __aenter__(self) -> "MsgHub":
        """Will be called when entering the MsgHub."""
        self._reset_subscriber()

        # broadcast the input message to all participants
        if self.announcement is not None:
            await self.broadcast(msg=self.announcement)

        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        """Will be called when exiting the MsgHub."""
        if self.enable_auto_broadcast:
            for agent in self.participants:
                agent.remove_subscribers(self.name)

    def _reset_subscriber(self) -> None:
        """Reset the subscriber for agent in `self.participant`"""
        if self.enable_auto_broadcast:
            for agent in self.participants:
                agent.reset_subscribers(self.name, self.participants)

    def add(
        self,
        new_participant: list[AgentBase] | AgentBase,
    ) -> None:
        """Add new participant into this hub"""
        if isinstance(new_participant, AgentBase):
            new_participant = [new_participant]

        for agent in new_participant:
            if agent not in self.participants:
                self.participants.append(agent)

        self._reset_subscriber()

    def delete(
        self,
        participant: list[AgentBase] | AgentBase,
    ) -> None:
        """Delete agents from participant."""
        if isinstance(participant, AgentBase):
            participant = [participant]

        for agent in participant:
            if agent in self.participants:
                # remove agent from self.participant
                self.participants.pop(self.participants.index(agent))
            else:
                logger.warning(
                    "Cannot find the agent with ID %s, skip its deletion.",
                    agent.id,
                )

        # Remove this agent from the subscriber of other agents
        self._reset_subscriber()

    async def broadcast(self, msg: list[Msg] | Msg) -> None:
        """Broadcast the message to all participants.

        Args:
            msg (`list[Msg] | Msg`):
                Message(s) to be broadcast among all participants.
        """
        for agent in self.participants:
            await agent.observe(msg)

    def set_auto_broadcast(self, enable: bool) -> None:
        """Enable automatic broadcasting of the replied message from any
        participant to all other participants.

        Args:
            enable (`bool`):
                Whether to enable automatic broadcasting. If disabled, the
                MsgHub will only serve as a manual message broadcaster with
                the `announcement` argument and the `broadcast()` method.
        """
        if enable:
            self.enable_auto_broadcast = True
            self._reset_subscriber()
        else:
            self.enable_auto_broadcast = False
            for agent in self.participants:
                agent.remove_subscribers(self.name)



# -*- coding: utf-8 -*-
"""A model class for RL Training with Trinity-RFT."""
from typing import (
    Optional,
    TYPE_CHECKING,
)
from ._openai_model import OpenAIChatModel
from ..types import JSONSerializableObject


if TYPE_CHECKING:
    from openai import AsyncOpenAI
else:
    AsyncOpenAI = "openai.AsyncOpenAI"


class TrinityChatModel(OpenAIChatModel):
    """A model class for RL Training with Trinity-RFT."""

    def __init__(
        self,
        openai_async_client: AsyncOpenAI,
        generate_kwargs: dict[str, JSONSerializableObject] | None = None,
        enable_thinking: Optional[bool] = None,
    ) -> None:
        """Initialize the Trinity model class.

        Args:
            openai_async_client (`AsyncOpenAI`):
                The OpenAI async client instance provided by Trinity-RFT.
            generate_kwargs (`dict[str, JSONSerializableObject] | None`, \
            optional):
                Additional keyword arguments to pass to the model's generate
                method. Defaults to None.
            enable_thinking (`bool`, optional):
                Whether to enable the model's thinking capability. Only
                applicable for Qwen3 series models. Defaults to None.
        """
        model_name = getattr(openai_async_client, "model_path", None)
        if model_name is None:
            raise ValueError(
                "The provided openai_async_client does not have a "
                "`model_path` attribute. Please ensure you are using "
                "the instance provided by Trinity-RFT.",
            )
        super().__init__(
            model_name=model_name,
            api_key="EMPTY",
            generate_kwargs=generate_kwargs,
            stream=False,  # RL training does not support streaming
        )
        if enable_thinking is not None:
            if "chat_template_kwargs" not in self.generate_kwargs:
                self.generate_kwargs["chat_template_kwargs"] = {}
            assert isinstance(
                self.generate_kwargs["chat_template_kwargs"],
                dict,
            ), "chat_template_kwargs must be a dictionary."
            self.generate_kwargs["chat_template_kwargs"][
                "enable_thinking"
            ] = enable_thinking
        # change the client instance to the provided one
        self.client = openai_async_client



# -*- coding: utf-8 -*-
"""The StdIO MCP server implementation in AgentScope, which provides
function-level fine-grained control over the MCP servers using standard IO."""
from typing import Literal

from mcp import stdio_client, StdioServerParameters

from ._stateful_client_base import StatefulClientBase


class StdIOStatefulClient(StatefulClientBase):
    """A client class that sets up and manage StdIO MCP server connections, and
    provides function-level fine-grained control over the MCP servers.

    .. tip:: The stateful client is recommended for MCP servers that need to
     maintain session states, e.g. web browsers or other interactive
     MCP servers.

    .. note:: The stateful client will maintain one session across multiple
     tool calls, until the client is closed by explicitly calling the
     `close()` method.

    .. note:: When multiple StdIOStatefulClient instances are connected,
     they should be closed following the Last In First Out (LIFO) principle
     to avoid potential errors. Always close the most recently registered
     client first, then work backwards to the first one.
     For more details, please refer to this `issue
     <https://github.com/modelcontextprotocol/python-sdk/issues/577>`_.
    """

    def __init__(
        self,
        name: str,
        command: str,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
        cwd: str | None = None,
        encoding: str = "utf-8",
        encoding_error_handler: Literal[
            "strict",
            "ignore",
            "replace",
        ] = "strict",
    ) -> None:
        """Initialize the MCP server with std IO.

        Args:
            name (`str`):
                The name to identify the MCP server, which should be unique
                across the MCP servers.
            command (`str`):
                The executable to run to start the server.
            args (`list[str] | None`, optional):
                Command line arguments to pass to the executable.
            env (`dict[str, str] | None`, optional):
                The environment to use when spawning the process.
            cwd (`str | None`, optional):
                The working directory to use when spawning the process.
            encoding (`str`, optional):
                The text encoding used when sending/receiving messages to the
                server. Defaults to "utf-8".
            encoding_error_handler (`Literal["strict", "ignore", "replace"]`, \
             defaults to "strict"):
                The text encoding error handler.
        """
        super().__init__(name=name)

        self.client = stdio_client(
            StdioServerParameters(
                command=command,
                args=args or [],
                env=env,
                cwd=cwd,
                encoding=encoding,
                encoding_error_handler=encoding_error_handler,
            ),
        )



# -*- coding: utf-8 -*-
"""The embedding cache base class."""
from abc import abstractmethod
from typing import List, Any

from ..types import (
    JSONSerializableObject,
    Embedding,
)


class EmbeddingCacheBase:
    """Base class for embedding caches, which is responsible for storing and
    retrieving embeddings."""

    @abstractmethod
    async def store(
        self,
        embeddings: List[Embedding],
        identifier: JSONSerializableObject,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """Store the embeddings with the given identifier.

        Args:
            embeddings (`List[Embedding]`):
                The embeddings to store.
            identifier (`JSONSerializableObject`):
                The identifier to distinguish the embeddings.
            overwrite (`bool`, defaults to `False`):
                Whether to overwrite existing embeddings with the same
                identifier. If `True`, existing embeddings will be replaced.
        """

    @abstractmethod
    async def retrieve(
        self,
        identifier: JSONSerializableObject,
    ) -> List[Embedding] | None:
        """Retrieve the embeddings with the given identifier. If not
        found, return `None`.

        Args:
            identifier (`JSONSerializableObject`):
                The identifier to retrieve the embeddings.
        """

    @abstractmethod
    async def remove(
        self,
        identifier: JSONSerializableObject,
    ) -> None:
        """Remove the embeddings with the given identifier.

        Args:
            identifier (`JSONSerializableObject`):
                The identifier to remove the embeddings.
        """

    @abstractmethod
    async def clear(self) -> None:
        """Clear all cached embeddings."""



# -*- coding: utf-8 -*-
"""Model wrapper for Ollama models."""
from datetime import datetime
from typing import (
    Any,
    TYPE_CHECKING,
    List,
    AsyncGenerator,
    AsyncIterator,
    Literal,
    Type,
)
from collections import OrderedDict

from pydantic import BaseModel

from . import ChatResponse
from ._model_base import ChatModelBase
from ._model_usage import ChatUsage
from .._logging import logger
from .._utils._common import _json_loads_with_repair
from ..message import ToolUseBlock, TextBlock, ThinkingBlock
from ..tracing import trace_llm


if TYPE_CHECKING:
    from ollama._types import ChatResponse as OllamaChatResponse
else:
    OllamaChatResponse = "ollama._types.ChatResponse"


class OllamaChatModel(ChatModelBase):
    """The Ollama chat model class in agentscope."""

    def __init__(
        self,
        model_name: str,
        stream: bool = False,
        options: dict = None,
        keep_alive: str = "5m",
        enable_thinking: bool | None = None,
        host: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Ollama chat model.

        Args:
           model_name (`str`):
               The name of the model.
           stream (`bool`, default `True`):
               Streaming mode or not.
           options (`dict`, default `None`):
               Additional parameters to pass to the Ollama API. These can
               include temperature etc.
           keep_alive (`str`, default `"5m"`):
               Duration to keep the model loaded in memory. The format is a
               number followed by a unit suffix (s for seconds, m for minutes
               , h for hours).
           enable_thinking (`bool | None`, default `None`)
               Whether enable thinking or not, only for models such as qwen3,
               deepseek-r1, etc. For more details, please refer to
               https://ollama.com/search?c=thinking
           host (`str | None`, default `None`):
               The host address of the Ollama server. If None, uses the
               default address (typically http://localhost:11434).
           **kwargs (`Any`):
               Additional keyword arguments to pass to the base chat model
               class.
        """

        try:
            import ollama
        except ImportError as e:
            raise ImportError(
                "The package ollama is not found. Please install it by "
                'running command `pip install "ollama>=0.1.7"`',
            ) from e

        super().__init__(model_name, stream)

        self.client = ollama.AsyncClient(
            host=host,
            **kwargs,
        )
        self.options = options
        self.keep_alive = keep_alive
        self.think = enable_thinking

    @trace_llm
    async def __call__(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict] | None = None,
        tool_choice: Literal["auto", "none", "any", "required"]
        | str
        | None = None,
        structured_model: Type[BaseModel] | None = None,
        **kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        """Get the response from Ollama chat completions API by the given
        arguments.

        Args:
            messages (`list[dict]`):
                A list of dictionaries, where `role` and `content` fields are
                required, and `name` field is optional.
            tools (`list[dict]`, default `None`):
                The tools JSON schemas that the model can use.
            tool_choice (`Literal["auto", "none", "any", "required"] | str \
                | None`, default `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", "any", "required", or specific tool
                 name.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.
            **kwargs (`Any`):
                The keyword arguments for Ollama chat completions API,
                e.g. `think`etc. Please refer to the Ollama API
                documentation for more details.

        Returns:
            `ChatResponse | AsyncGenerator[ChatResponse, None]`:
                The response from the Ollama chat completions API.
        """

        kwargs = {
            "model": self.model_name,
            "messages": messages,
            "stream": self.stream,
            "options": self.options,
            "keep_alive": self.keep_alive,
            **kwargs,
        }

        if self.think is not None and "think" not in kwargs:
            kwargs["think"] = self.think

        if tools:
            kwargs["tools"] = self._format_tools_json_schemas(tools)

        if tool_choice:
            logger.warning("Ollama does not support tool_choice yet, ignored.")

        if structured_model:
            kwargs["format"] = structured_model.model_json_schema()

        start_datetime = datetime.now()
        response = await self.client.chat(**kwargs)

        if self.stream:
            return self._parse_ollama_stream_completion_response(
                start_datetime,
                response,
                structured_model,
            )

        parsed_response = await self._parse_ollama_completion_response(
            start_datetime,
            response,
            structured_model,
        )

        return parsed_response

    async def _parse_ollama_stream_completion_response(
        self,
        start_datetime: datetime,
        response: AsyncIterator[OllamaChatResponse],
        structured_model: Type[BaseModel] | None = None,
    ) -> AsyncGenerator[ChatResponse, None]:
        """Given an Ollama streaming completion response, extract the
        content blocks and usages from it and yield ChatResponse objects.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`AsyncIterator[OllamaChatResponse]`):
                Ollama streaming response async iterator to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            AsyncGenerator[ChatResponse, None]:
                An async generator that yields ChatResponse objects containing
                the content blocks and usage information for each chunk in the
                streaming response.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.

        """
        accumulated_text = ""
        acc_thinking_content = ""
        tool_calls = OrderedDict()  # Store tool calls
        metadata: dict | None = None

        async for chunk in response:
            # Handle text content
            msg = chunk.message
            acc_thinking_content += msg.thinking or ""
            accumulated_text += msg.content or ""

            # Handle tool calls
            for idx, tool_call in enumerate(msg.tool_calls or []):
                function = tool_call.function
                tool_id = f"{idx}_{function.name}"
                tool_calls[tool_id] = {
                    "type": "tool_use",
                    "id": tool_id,
                    "name": function.name,
                    "input": function.arguments,
                }
            # Calculate usage statistics
            current_time = (datetime.now() - start_datetime).total_seconds()
            usage = ChatUsage(
                input_tokens=getattr(chunk, "prompt_eval_count", 0) or 0,
                output_tokens=getattr(chunk, "eval_count", 0) or 0,
                time=current_time,
            )
            # Create content blocks
            contents: list = []

            if acc_thinking_content:
                contents.append(
                    ThinkingBlock(
                        type="thinking",
                        thinking=acc_thinking_content,
                    ),
                )

            if accumulated_text:
                contents.append(TextBlock(type="text", text=accumulated_text))
                if structured_model:
                    metadata = _json_loads_with_repair(accumulated_text)

            # Add tool call blocks
            for tool_call in tool_calls.values():
                try:
                    input_data = tool_call["input"]
                    if isinstance(input_data, str):
                        input_data = _json_loads_with_repair(input_data)
                    contents.append(
                        ToolUseBlock(
                            type=tool_call["type"],
                            id=tool_call["id"],
                            name=tool_call["name"],
                            input=input_data,
                        ),
                    )
                except Exception as e:
                    print(f"Error parsing tool call input: {e}")

            # Generate response when there's new content or at final chunk
            if chunk.done and contents:
                res = ChatResponse(
                    content=contents,
                    usage=usage,
                    metadata=metadata,
                )
                yield res

    async def _parse_ollama_completion_response(
        self,
        start_datetime: datetime,
        response: OllamaChatResponse,
        structured_model: Type[BaseModel] | None = None,
    ) -> ChatResponse:
        """Given an Ollama chat completion response object, extract the content
        blocks and usages from it.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`OllamaChatResponse`):
                Ollama OllamaChatResponse object to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            `ChatResponse`:
                A ChatResponse object containing the content blocks and usage.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        content_blocks: List[TextBlock | ToolUseBlock | ThinkingBlock] = []
        metadata: dict | None = None

        if response.message.thinking:
            content_blocks.append(
                ThinkingBlock(
                    type="thinking",
                    thinking=response.message.thinking,
                ),
            )

        if response.message.content:
            content_blocks.append(
                TextBlock(
                    type="text",
                    text=response.message.content,
                ),
            )
            if structured_model:
                metadata = _json_loads_with_repair(
                    response.message.content,
                )

        for idx, tool_call in enumerate(response.message.tool_calls or []):
            content_blocks.append(
                ToolUseBlock(
                    type="tool_use",
                    id=f"{idx}_{tool_call.function.name}",
                    name=tool_call.function.name,
                    input=tool_call.function.arguments,
                ),
            )

        usage = None
        if "prompt_eval_count" in response and "eval_count" in response:
            usage = ChatUsage(
                input_tokens=response.get("prompt_eval_count", 0),
                output_tokens=response.get("eval_count", 0),
                time=(datetime.now() - start_datetime).total_seconds(),
            )

        parsed_response = ChatResponse(
            content=content_blocks,
            usage=usage,
            metadata=metadata,
        )

        return parsed_response

    def _format_tools_json_schemas(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schemas to the Ollama format."""
        return schemas



# -*- coding: utf-8 -*-
# mypy: disable-error-code="dict-item"
"""The Google Gemini model in agentscope."""
from datetime import datetime
from typing import (
    AsyncGenerator,
    Any,
    TYPE_CHECKING,
    AsyncIterator,
    Literal,
    Type,
    List,
)

from pydantic import BaseModel

from .._logging import logger
from .._utils._common import _json_loads_with_repair
from ..message import ToolUseBlock, TextBlock, ThinkingBlock
from ._model_usage import ChatUsage
from ._model_base import ChatModelBase
from ._model_response import ChatResponse
from ..tracing import trace_llm
from ..types import JSONSerializableObject

if TYPE_CHECKING:
    from google.genai.types import GenerateContentResponse
else:
    GenerateContentResponse = "google.genai.types.GenerateContentResponse"


class GeminiChatModel(ChatModelBase):
    """The Google Gemini chat model class in agentscope."""

    def __init__(
        self,
        model_name: str,
        api_key: str,
        stream: bool = True,
        thinking_config: dict | None = None,
        client_args: dict = None,
        generate_kwargs: dict[str, JSONSerializableObject] | None = None,
    ) -> None:
        """Initialize the Gemini chat model.

        Args:
            model_name (`str`):
                The name of the Gemini model to use, e.g. "gemini-2.5-flash".
            api_key (`str`):
                The API key for Google Gemini.
            stream (`bool`, default `True`):
                Whether to use streaming output or not.
            thinking_config (`dict | None`, optional):
                Thinking config, supported models are 2.5 Pro, 2.5 Flash, etc.
                Refer to https://ai.google.dev/gemini-api/docs/thinking for
                more details.

                .. code-block:: python
                    :caption: Example of thinking_config

                    {
                        "include_thoughts": True, # enable thoughts or not
                        "thinking_budget": 1024   # Max tokens for reasoning
                    }

            client_args (`dict`, default `None`):
                The extra keyword arguments to initialize the OpenAI client.
            generate_kwargs (`dict[str, JSONSerializableObject] | None`, \
             optional):
               The extra keyword arguments used in Gemini API generation,
               e.g. `temperature`, `seed`.
        """
        try:
            from google import genai
        except ImportError as e:
            raise ImportError(
                "Please install gemini Python sdk with "
                "`pip install -q -U google-genai`",
            ) from e

        super().__init__(model_name, stream)

        self.client = genai.Client(
            api_key=api_key,
            **(client_args or {}),
        )
        self.thinking_config = thinking_config
        self.generate_kwargs = generate_kwargs or {}

    @trace_llm
    async def __call__(
        self,
        messages: list[dict],
        tools: list[dict] | None = None,
        tool_choice: Literal["auto", "none", "any", "required"]
        | str
        | None = None,
        structured_model: Type[BaseModel] | None = None,
        **config_kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        """Call the Gemini model with the provided arguments.

        Args:
            messages (`list[dict[str, Any]]`):
                A list of dictionaries, where `role` and `content` fields are
                required.
            tools (`list[dict] | None`, default `None`):
                The tools JSON schemas that the model can use.
            tool_choice (`Literal["auto", "none", "any", "required"] | str \
            | None`, default `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", "any", "required", or specific tool
                 name. For more details, please refer to
                 https://ai.google.dev/gemini-api/docs/function-calling?hl=en&example=meeting#function_calling_modes
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

                .. note:: When `structured_model` is specified,
                    both `tools` and `tool_choice` parameters are ignored,
                    and the model will only perform structured output
                    generation without calling any other tools.

                For more details, please refer to
                    https://ai.google.dev/gemini-api/docs/structured-output

            **config_kwargs (`Any`):
                The keyword arguments for Gemini chat completions API.
        """

        config: dict = {
            "thinking_config": self.thinking_config,
            **self.generate_kwargs,
            **config_kwargs,
        }

        if tools:
            config["tools"] = self._format_tools_json_schemas(tools)

        if tool_choice:
            self._validate_tool_choice(tool_choice, tools)
            config["tool_config"] = self._format_tool_choice(tool_choice)

        if structured_model:
            if tools or tool_choice:
                logger.warning(
                    "structured_model is provided. Both 'tools' and "
                    "'tool_choice' parameters will be overridden and "
                    "ignored. The model will only perform structured output "
                    "generation without calling any other tools.",
                )
            config.pop("tools", None)
            config.pop("tool_config", None)
            config["response_mime_type"] = "application/json"
            config["response_schema"] = structured_model

        # Prepare the arguments for the Gemini API call
        kwargs: dict[str, JSONSerializableObject] = {
            "model": self.model_name,
            "contents": messages,
            "config": config,
        }

        start_datetime = datetime.now()
        if self.stream:
            response = await self.client.aio.models.generate_content_stream(
                **kwargs,
            )

            return self._parse_gemini_stream_generation_response(
                start_datetime,
                response,
                structured_model,
            )

        # non-streaming
        response = await self.client.aio.models.generate_content(
            **kwargs,
        )

        parsed_response = self._parse_gemini_generation_response(
            start_datetime,
            response,
            structured_model,
        )

        return parsed_response

    async def _parse_gemini_stream_generation_response(
        self,
        start_datetime: datetime,
        response: AsyncIterator[GenerateContentResponse],
        structured_model: Type[BaseModel] | None = None,
    ) -> AsyncGenerator[ChatResponse, None]:
        """Given a Gemini streaming generation response, extract the
        content blocks and usages from it and yield ChatResponse objects.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`AsyncIterator[GenerateContentResponse]`):
                Gemini GenerateContentResponse async iterator to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            `AsyncGenerator[ChatResponse, None]`:
                An async generator that yields ChatResponse objects containing
                the content blocks and usage information for each chunk in the
                streaming response.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """

        text = ""
        thinking = ""
        metadata: dict | None = None
        async for chunk in response:
            content_block: list = []

            # Thinking parts
            if (
                chunk.candidates
                and chunk.candidates[0].content
                and chunk.candidates[0].content.parts
            ):
                for part in chunk.candidates[0].content.parts:
                    if part.thought and part.text:
                        thinking += part.text

            # Text parts
            if chunk.text:
                text += chunk.text
                if structured_model:
                    metadata = _json_loads_with_repair(text)

            # Function calls
            tool_calls = []
            if chunk.function_calls:
                for function_call in chunk.function_calls:
                    tool_calls.append(
                        ToolUseBlock(
                            type="tool_use",
                            id=function_call.id,
                            name=function_call.name,
                            input=function_call.args or {},
                        ),
                    )

            usage = None
            if chunk.usage_metadata:
                usage = ChatUsage(
                    input_tokens=chunk.usage_metadata.prompt_token_count,
                    output_tokens=chunk.usage_metadata.total_token_count
                    - chunk.usage_metadata.prompt_token_count,
                    time=(datetime.now() - start_datetime).total_seconds(),
                )

            if thinking:
                content_block.append(
                    ThinkingBlock(
                        type="thinking",
                        thinking=thinking,
                    ),
                )

            if text:
                content_block.append(
                    TextBlock(
                        type="text",
                        text=text,
                    ),
                )

            content_block.extend(
                [
                    *tool_calls,
                ],
            )

            parsed_chunk = ChatResponse(
                content=content_block,
                usage=usage,
                metadata=metadata,
            )
            yield parsed_chunk

    def _parse_gemini_generation_response(
        self,
        start_datetime: datetime,
        response: GenerateContentResponse,
        structured_model: Type[BaseModel] | None = None,
    ) -> ChatResponse:
        """Given a Gemini chat completion response object, extract the content
           blocks and usages from it.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (`ChatCompletion`):
                The OpenAI chat completion response object to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            ChatResponse (`ChatResponse`):
                A ChatResponse object containing the content blocks and usage.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        content_blocks: List[TextBlock | ToolUseBlock | ThinkingBlock] = []
        metadata: dict | None = None

        if (
            response.candidates
            and response.candidates[0].content
            and response.candidates[0].content.parts
        ):
            for part in response.candidates[0].content.parts:
                if part.thought and part.text:
                    content_blocks.append(
                        ThinkingBlock(
                            type="thinking",
                            thinking=part.text,
                        ),
                    )

        if response.text:
            content_blocks.append(
                TextBlock(
                    type="text",
                    text=response.text,
                ),
            )
            if structured_model:
                metadata = _json_loads_with_repair(response.text)

        if response.function_calls:
            for tool_call in response.function_calls:
                content_blocks.append(
                    ToolUseBlock(
                        type="tool_use",
                        id=tool_call.id,
                        name=tool_call.name,
                        input=tool_call.args or {},
                    ),
                )

        if response.usage_metadata:
            usage = ChatUsage(
                input_tokens=response.usage_metadata.prompt_token_count,
                output_tokens=response.usage_metadata.total_token_count
                - response.usage_metadata.prompt_token_count,
                time=(datetime.now() - start_datetime).total_seconds(),
            )

        else:
            usage = None

        return ChatResponse(
            content=content_blocks,
            usage=usage,
            metadata=metadata,
        )

    def _format_tools_json_schemas(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schema into required format for Gemini API.

        Args:
            schemas (`dict[str, Any]`):
                The tools JSON schemas.

        Returns:
            List[Dict[str, Any]]:
                A list containing a dictionary with the
                "function_declarations" key, which maps to a list of
                function definitions.

        Example:
            .. code-block:: python
                :caption: Example tool schemas of Gemini API

                # Input JSON schema
                schemas = [
                    {
                        'type': 'function',
                        'function': {
                            'name': 'execute_shell_command',
                            'description': 'xxx',
                            'parameters': {
                                'type': 'object',
                                'properties': {
                                    'command': {
                                        'type': 'string',
                                        'description': 'xxx.'
                                    },
                                    'timeout': {
                                        'type': 'integer',
                                        'default': 300
                                    }
                                },
                                'required': ['command']
                            }
                        }
                    }
                ]

                # Output format (Gemini API expected):
                [
                    {
                        'function_declarations': [
                            {
                                'name': 'execute_shell_command',
                                'description': 'xxx.',
                                'parameters': {
                                    'type': 'object',
                                    'properties': {
                                        'command': {
                                            'type': 'string',
                                            'description': 'xxx.'
                                        },
                                        'timeout': {
                                            'type': 'integer',
                                            'default': 300
                                        }
                                    },
                                    'required': ['command']
                                }
                            }
                        ]
                    }
                ]
        """
        return [
            {
                "function_declarations": [
                    _["function"] for _ in schemas if "function" in _
                ],
            },
        ]

    def _format_tool_choice(
        self,
        tool_choice: Literal["auto", "none", "any", "required"] | str | None,
    ) -> dict | None:
        """Format tool_choice parameter for API compatibility.

        Args:
            tool_choice (`Literal["auto", "none"] | str | None`, default \
            `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", "any", "required", or specific tool
                 name.
                 For more details, please refer to
                 https://ai.google.dev/gemini-api/docs/function-calling?hl=en&example=meeting#function_calling_modes
        Returns:
            `dict | None`:
                The formatted tool choice configuration dict, or None if
                    tool_choice is None.
        """
        if tool_choice is None:
            return None

        mode_mapping = {
            "auto": "AUTO",
            "none": "NONE",
            "any": "ANY",
            "required": "ANY",
        }
        mode = mode_mapping.get(tool_choice)
        if mode:
            return {"function_calling_config": {"mode": mode}}
        return {
            "function_calling_config": {
                "mode": "ANY",
                "allowed_function_names": [tool_choice],
            },
        }



# -*- coding: utf-8 -*-
"""The ollama text embedding model class."""
from datetime import datetime
from typing import List, Any

from ._embedding_response import EmbeddingResponse
from ._embedding_usage import EmbeddingUsage
from ._cache_base import EmbeddingCacheBase
from ..embedding import EmbeddingModelBase
from ..message import TextBlock


class OllamaTextEmbedding(EmbeddingModelBase):
    """The Ollama embedding model."""

    supported_modalities: list[str] = ["text"]
    """This class only supports text input."""

    def __init__(
        self,
        model_name: str,
        dimensions: int,
        host: str | None = None,
        embedding_cache: EmbeddingCacheBase | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Ollama text embedding model class.

        Args:
            model_name (`str`):
                The name of the embedding model.
            dimensions (`int`):
                The dimension of the embedding vector, the parameter should be
                provided according to the model used.
            host (`str | None`, defaults to `None`):
                The host URL for the Ollama API.
            embedding_cache (`EmbeddingCacheBase | None`, defaults to `None`):
                The embedding cache class instance, used to cache the
                embedding results to avoid repeated API calls.
        """
        import ollama

        super().__init__(model_name, dimensions)

        self.client = ollama.AsyncClient(host=host, **kwargs)
        self.embedding_cache = embedding_cache

    async def __call__(
        self,
        text: List[str | TextBlock],
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """Call the Ollama embedding API.

        Args:
            text (`List[str | TextBlock]`):
                The input text to be embedded. It can be a list of strings.
        """
        gather_text = []
        for _ in text:
            if isinstance(_, dict) and "text" in _:
                gather_text.append(_["text"])
            elif isinstance(_, str):
                gather_text.append(_)
            else:
                raise ValueError(
                    "Input text must be a list of strings or TextBlock dicts.",
                )

        kwargs = {
            "input": gather_text,
            "model": self.model_name,
            "dimensions": self.dimensions,
            **kwargs,
        }

        if self.embedding_cache:
            cached_embeddings = await self.embedding_cache.retrieve(
                identifier=kwargs,
            )
            if cached_embeddings:
                return EmbeddingResponse(
                    embeddings=cached_embeddings,
                    usage=EmbeddingUsage(
                        tokens=0,
                        time=0,
                    ),
                    source="cache",
                )

        start_time = datetime.now()
        response = await self.client.embed(**kwargs)
        time = (datetime.now() - start_time).total_seconds()

        if self.embedding_cache:
            await self.embedding_cache.store(
                identifier=kwargs,
                embeddings=response.embeddings,
            )

        return EmbeddingResponse(
            embeddings=response.embeddings,
            usage=EmbeddingUsage(
                time=time,
            ),
        )



# -*- coding: utf-8 -*-
# pylint: disable=R0901
"""The content blocks of messages"""

from typing import Literal, List
from typing_extensions import TypedDict, Required


class TextBlock(TypedDict, total=False):
    """The text block."""

    type: Required[Literal["text"]]
    """The type of the block"""
    text: str
    """The text content"""


class ThinkingBlock(TypedDict, total=False):
    """The thinking block."""

    type: Required[Literal["thinking"]]
    """The type of the block"""
    thinking: str


class Base64Source(TypedDict, total=False):
    """The base64 source"""

    type: Required[Literal["base64"]]
    """The type of the src, must be `base64`"""

    media_type: Required[str]
    """The media type of the data, e.g. `image/jpeg` or `audio/mpeg`"""

    data: Required[str]
    """The base64 data, in format of RFC 2397"""


class URLSource(TypedDict, total=False):
    """The URL source"""

    type: Required[Literal["url"]]
    """The type of the src, must be `url`"""

    url: Required[str]
    """The URL of the image or audio"""


class ImageBlock(TypedDict, total=False):
    """The image block"""

    type: Required[Literal["image"]]
    """The type of the block"""

    source: Required[Base64Source | URLSource]
    """The src of the image"""


class AudioBlock(TypedDict, total=False):
    """The audio block"""

    type: Required[Literal["audio"]]
    """The type of the block"""

    source: Required[Base64Source | URLSource]
    """The src of the audio"""


class VideoBlock(TypedDict, total=False):
    """The video block"""

    type: Required[Literal["video"]]
    """The type of the block"""

    source: Required[Base64Source | URLSource]
    """The src of the audio"""


class ToolUseBlock(TypedDict, total=False):
    """The tool use block."""

    type: Required[Literal["tool_use"]]
    """The type of the block, must be `tool_use`"""
    id: Required[str]
    """The identity of the tool call"""
    name: Required[str]
    """The name of the tool"""
    input: Required[dict[str, object]]
    """The input of the tool"""


class ToolResultBlock(TypedDict, total=False):
    """The tool result block."""

    type: Required[Literal["tool_result"]]
    """The type of the block"""
    id: Required[str]
    """The identity of the tool call result"""
    output: Required[str | List[TextBlock | ImageBlock | AudioBlock]]
    """The output of the tool function"""
    name: Required[str]
    """The name of the tool function"""


# The content block
ContentBlock = (
    ToolUseBlock
    | ToolResultBlock
    | TextBlock
    | ThinkingBlock
    | ImageBlock
    | AudioBlock
    | VideoBlock
)
